/*
 This file is part of QxBF.

 QxBF, a preprocessor for QBF based on failed literal detection.	
 Copyright 2011 Florian Lonsing, Aina Niemetz and Mathias Preiner.

 See also http://fmv.jku.at/qxbf for further information.

 QxBF is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or (at
 your option) any later version.

 QxBF is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with QxBF.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <sys/resource.h>
#include <stdlib.h>
#include <error.h>
#include <unistd.h>
#include <stdio.h>
#include <signal.h>
#include <ctype.h>
#include <string.h>
#include <dirent.h>
#include <time.h>
#include <assert.h>
#include "qxbf.h"
#include "../picosat/picosat.h"

#define VERSION								\
  "QxBF 1.1\n"							\
  "Copyright 2011 Florian Lonsing, Aina Niemetz and Mathias Preiner.\n"\
  "This is free software; see COPYING for copying conditions.\n"	\
  "There is NO WARRANTY, to the extent permitted by law.\n"

/* Get random value in range [low,high]. */
#define GET_RAND(low,high) ((rand() % (high-low+1))+low)

/* Get scope-type of current variable wrt. current abstraction of scopes. */
#define ABS_VAR_IS_EXISTS(var)						\
  (vars[(var)].type == QTYPE_EXISTS ||					\
   (decision_var > 0 && vars[(var)].type == QTYPE_FORALL &&		\
    vars[(var)].s_level < vars[decision_var].s_level))

static FILE *in = NULL;
static char *in_filename = 0;

static int num_vars = 0;
static int num_clauses = 0;
static int num_scopes = 0;

static Var *vars = NULL;
static Clause *clauses = NULL;
static Lit *literals = NULL;
static Scope *scopes = NULL;

/* -1 if formula is not closed and an additional existential scope has been
 * added as the outermost scope. */
static int outermost_scope_index = 0;

static VarId *scope_vars = NULL;
static ClauseId *occurences = NULL;

static double start_time = 0;
static double end_time = 0;

static double cur_bcp_time_stamp = 0;
static double cur_sat_time_stamp = 0;

static int computed_dec_limit = 0;

static Stack assigned_vars;
static int *assigned_vars_l0 = NULL;    /* decision level 0 */
static int *assigned_vars_l1 = NULL;    /* decision level 1 */
static int scope_var_index = 0; /* next index of var in scope to be used 
                                   as decision */

static int *next_to_propagate = 0;
static int scope_index = 0;
static VarId decision_var = 0;

static QxBFOptions options;

/* Flag to check if one phase has been tried. */
static unsigned int tried = 0;
static unsigned int bcp_rounds = 0;
static unsigned int sat_rounds = 0;
static unsigned int total_ignore_literals = 0;

/* Propagation bounds. BCP can do at most 'num_vars' propagations, so
   internally 'max_prop_bound' will not be larger than that value. */
static unsigned int max_bcp_prop_bound = 0;
static unsigned int cur_bcp_prop_bound = 0;
/* Largest value of 'cur_prop_bound' where changes of top-level
   assignment occurred. This is an upper bound on the optimal number of
   propagations needed to identify all possible units. */
static unsigned int opt_cur_bcp_prop_bound = 0;

static unsigned int max_sat_prop_bound = 0;
static unsigned int cur_sat_prop_bound = 0;
static unsigned int opt_cur_sat_prop_bound = 0;
/* Flag for indicating whether quantifier-free CNF is already UNSAT. */
static unsigned int cnf_unsat = 0;

static unsigned int learnt_from_unsat_tests = 0;
static unsigned int learnt_from_sat_solver = 0;
static unsigned int total_sat_solver_models = 0;

/* Number of propagations of current decision. */
static unsigned int cur_bcp_propagations = 0;
/* Largest value of 'cur_propagations' ever encountered. */
static unsigned int max_cur_bcp_propagations = 0;

/* Units learnt from failed literals. */
static unsigned int learnt_from_failed_lits = 0;
static unsigned long long int total_bcp_propagations = 0;
static unsigned long long int total_bcp_decisions = 0;

static unsigned long long int total_sat_tests = 0;
static double total_sat_time = 0;
static double total_bcp_time = 0;

/* Data used for Q-resolution. */
static Stack qres_wreason = { 0, 0, 0 };
static Stack qres_antecedent = { 0, 0, 0 };
static Stack qres_tmp = { 0, 0, 0 };
static ClauseId empty_clause = INVALID_CLAUSE_ID;
static int qres_derived_empty_clause = 0;


static double
time_stamp ()
{
  double result = 0;
  struct rusage usage;

  if (!getrusage (RUSAGE_SELF, &usage))
    {
      result += usage.ru_utime.tv_sec + 1e-6 * usage.ru_utime.tv_usec;
      result += usage.ru_stime.tv_sec + 1e-6 * usage.ru_stime.tv_usec;
    }

  return result;
}


/*
 * Parse a QBF in QDIMACS format. 
 * Scopes and clauses may come in no particular order, given number of 
 * variables/clauses may serve as min values. 
 * Given formula may contain free variables. 
 */
static QxBFState
parse (void)
{
  assert (in != NULL);

  char c;
  int i, num, level, *tmp, neg = 0, cnt_clauses = 0, cnt_literals = 0;
  QType first_scope_type = QTYPE_UNDEF;
  QType scope_type = QTYPE_UNDEF;
  Stack stack, stack_free_vars;
  ClauseId *occs_tmp;
  Lit *lits_tmp, unit;
  VarId *scope_vars_tmp;
  BType assignment;
  QxBFState result = STATE_UNDEF;

  PARSER_SKIP_SPACE_DO_WHILE (c);

  QXBF_ABORT (c == EOF, "invalid input (empty file)");

  while (c == QDIMACS_COMMENT)
    {
      while ((c = getc (in)) != '\n' && c != EOF);
      PARSER_SKIP_SPACE_DO_WHILE (c);
    }

  PARSER_SKIP_SPACE_WHILE (c);
  QXBF_ABORT (c != QDIMACS_PREAMBLE, "missing preamble");

  /* preamble */
  PARSER_SKIP_SPACE_DO_WHILE (c);
  QXBF_ABORT (c != QDIMACS_C, "invalid preamble");
  PARSER_SKIP_SPACE_DO_WHILE (c);
  QXBF_ABORT (c != QDIMACS_N, "invalid preamble");
  PARSER_SKIP_SPACE_DO_WHILE (c);
  QXBF_ABORT (c != QDIMACS_F, "invalid preamble");
  PARSER_SKIP_SPACE_DO_WHILE (c);

  /* number of variables */
  PARSER_READ_NUM (num_vars, c, "invalid preamble");
  vars = (Var *) malloc ((num_vars + 1) * sizeof (Var));
  QXBF_ABORT (vars == NULL, "memory allocation failed");
  memset (vars, 0, (num_vars + 1) * sizeof (Var));
  PARSER_SKIP_SPACE_WHILE (c);

  /* number of clauses */
  PARSER_READ_NUM (num_clauses, c, "invalid preamble");
  clauses = (Clause *) malloc (num_clauses * sizeof (Clause));
  QXBF_ABORT (clauses == NULL, "memory allocation failed");
  memset (clauses, 0, num_clauses * sizeof (Clause));
  PARSER_SKIP_SPACE_WHILE (c);

  assert (!isspace (c));

  STACK_INIT (stack, PARSER_STACKSIZE);
  STACK_INIT (stack_free_vars, PARSER_STACKSIZE);

  /* scopes / clauses */
  for (;;)
    {
      if (c == QDIMACS_EXISTS || c == QDIMACS_FORALL)
        {
          QXBF_ABORT (cnt_clauses > 0,
                      "clause expected, found quantifier set");

          PUSH (stack, EOL);    /* scope marker */
          num_scopes += 1;


          if (c == QDIMACS_EXISTS)
            scope_type = QTYPE_EXISTS;
          else
            scope_type = QTYPE_FORALL;
          PUSH (stack, scope_type);

          if (first_scope_type == QTYPE_UNDEF)
            first_scope_type = scope_type;

          for (;;)
            {
              PARSER_SKIP_SPACE_DO_WHILE (c);

              if (c == QDIMACS_EOL)
                {
                  PUSH (stack, EOL);
                  break;
                }

              if (c == EOF)
                goto FINALIZE_DATA_STRUCTURES;

              PARSER_READ_NUM (num, c, "digit expected");

              if (num > num_vars)
                {
                  vars = (Var *) realloc (vars, (num + 1) * sizeof (Var));
                  QXBF_ABORT (vars == NULL, "memory reallocation failed");
                  memset (vars + num_vars + 1, 0,
                          (num - num_vars) * sizeof (Var));
                  num_vars = num;
                }

              vars[num].id = (VarId) num;
              vars[num].type = scope_type;
              vars[num].mode = VMODE_UNDEF;
              vars[num].d_level = DLEVEL_UNDEF;
              vars[num].s_level = num_scopes - 1;
              vars[num].assignment = BTYPE_UNDEF;
              vars[num].is_propagated = 0;
              vars[num].num_neg_occs = 0;
              vars[num].num_pos_occs = 0;
              vars[num].num_neg_occs_cur = 0;
              vars[num].num_pos_occs_cur = 0;

              PUSH (stack, num);
            }
        }
      else
        {
          if (cnt_clauses >= num_clauses)
            {
              clauses =
                (Clause *) realloc (clauses,
                                    (cnt_clauses + 1) * sizeof (Clause));
              QXBF_ABORT (clauses == NULL, "memory reallocation failed");
              num_clauses = cnt_clauses;
              clauses[cnt_clauses].num_lits = 0;
            }
          clauses[cnt_clauses].id = cnt_clauses;
          clauses[cnt_clauses].sat_var = 0;
          clauses[cnt_clauses].is_pure_sat = 0;
          clauses[cnt_clauses].is_tautology = 0;

          for (;;)
            {
              if (c == QDIMACS_EOL)
                {
                  PUSH (stack, EOL);
                  break;
                }

              if (c == EOF)
                goto FINALIZE_DATA_STRUCTURES;

              if (c == MINUS)
                {
                  neg = 1;
                  c = getc (in);
                }

              PARSER_READ_NUM (num, c, "digit expected");

              /* check for free vars */
              if (num > num_vars)
                {
                  vars = (Var *) realloc (vars, (num + 1) * sizeof (Var));
                  QXBF_ABORT (vars == NULL, "memory reallocation failed");
                  memset (vars + num_vars + 1, 0,
                          (num - num_vars) * sizeof (Var));
                  num_vars = num;
                }

              /* add outermost existential scope for free variables */
              if (vars[num].type == QTYPE_UNDEF)
                {
                  if (first_scope_type == QTYPE_FORALL)
                    outermost_scope_index = -1;

                  vars[num].id = (VarId) num;
                  vars[num].type = QTYPE_EXISTS;
                  vars[num].mode = VMODE_UNDEF;
                  vars[num].d_level = DLEVEL_UNDEF;
                  vars[num].s_level = outermost_scope_index;
                  vars[num].assignment = BTYPE_UNDEF;
                  PUSH (stack_free_vars, num);
                }

              if (neg)
                {
                  vars[num].num_neg_occs += 1;
                  num = -num;
                }
              else
                vars[num].num_pos_occs += 1;

              PUSH (stack, num);
              cnt_literals += 1;
              clauses[cnt_clauses].num_lits += 1;
              neg = 0;

              PARSER_SKIP_SPACE_DO_WHILE (c);
            }
          cnt_clauses += 1;
        }
      PARSER_SKIP_SPACE_DO_WHILE (c);
    }

FINALIZE_DATA_STRUCTURES:
  literals = (Lit *) malloc (cnt_literals * sizeof (Lit));
  QXBF_ABORT (literals == NULL, "memory allocation failed");
  occurences = (ClauseId *) malloc (cnt_literals * sizeof (ClauseId));
  QXBF_ABORT (occurences == NULL, "memory allocation failed");

  if (outermost_scope_index == -1 ||
      (stack_free_vars.bp != stack_free_vars.sp && num_scopes < 1))
    {
      /* create additional outermost existential scope */
      scopes = (Scope *) malloc ((num_scopes + 1) * sizeof (Scope));
      QXBF_ABORT (scopes == NULL, "memory allocation failed");
      memset (scopes, 0, (num_scopes + 1) * sizeof (Scope));
      if (outermost_scope_index == -1)
        scopes += 1;
    }
  else
    {
      scopes = (Scope *) malloc (num_scopes * sizeof (Scope));
      QXBF_ABORT (scopes == NULL, "memory allocation failed");
      memset (scopes, 0, num_scopes * sizeof (Scope));
    }

  scope_vars = (VarId *) malloc (num_vars * sizeof (VarId));
  QXBF_ABORT (scope_vars == NULL, "memory allocation failed");
  scope_index = outermost_scope_index;

  STACK_INIT (assigned_vars, (num_vars + 1));
  assigned_vars_l0 = assigned_vars.bp;
  assigned_vars_l1 = assigned_vars.bp;

  /* set pointer to occurences (must be done explicitely in case of variables
   * with undefined scope)  */
  occs_tmp = occurences;
  for (i = 1; i <= num_vars; i++)
    {
      vars[i].sat_test_pos_ignore = vars[i].sat_test_neg_ignore =
        vars[i].learnt_from_unsat_test = 0;
      vars[i].antecedent = INVALID_CLAUSE_ID;
      vars[i].assignment = BTYPE_UNDEF;
      vars[i].trail_pos = INVALID_TRAIL_POS;
      vars[i].num_pos_occs_cur = vars[i].num_pos_occs;
      vars[i].num_neg_occs_cur = vars[i].num_neg_occs;

      vars[i].neg_occs = occs_tmp;
      occs_tmp += vars[i].num_neg_occs;
      vars[i].pos_occs = occs_tmp;
      occs_tmp += vars[i].num_pos_occs;
      vars[i].num_neg_occs = 0;
      vars[i].num_pos_occs = 0;
    }

  scope_vars_tmp = scope_vars;
  tmp = stack_free_vars.bp;
  level = outermost_scope_index;

  /* add free variables to outermost scope */
  if (tmp != stack_free_vars.sp)
    {
      scopes[level].type = QTYPE_EXISTS;
      scopes[level].level = level;
      scopes[level].vars = scope_vars_tmp;

      for (; tmp != stack_free_vars.sp; tmp++)
        scopes[level].vars[scopes[level].num_vars++] = *tmp;

      if (level == -1)
        scope_vars_tmp += scopes[level].num_vars;
    }

  /* add bound variables to their respective scopes */
  lits_tmp = literals;
  tmp = stack.bp;
  level = 0;
  cnt_clauses = 0;

  for (;;)
    {
      if (tmp == stack.sp)
        break;

      if (level < num_scopes && *tmp == EOL)    /* scope */
        {
          tmp += 1;

          scopes[level].type = (QType) * tmp;
          scopes[level].level = level;
          scopes[level].vars = scope_vars_tmp;

          for (tmp += 1; *tmp != EOL; tmp++)
            scopes[level].vars[scopes[level].num_vars++] = *tmp;

          scope_vars_tmp += scopes[level].num_vars;
          assert (*tmp == EOL);
          tmp += 1;
          level += 1;
        }
      else                      /* clause */
        {
          int tautology = 0;
          /* Prevent complementary clauses. */
          int *t1;
          for (t1 = tmp; *t1 != EOL && !tautology; t1++)
            {
              int *t2;
              for (t2 = t1 + 1; *t2 != EOL && !tautology; t2++)
                {
                  if (*t1 == -(*t2))
                    tautology = 1;
                }
            }

          clauses[cnt_clauses].is_tautology = tautology;
          clauses[cnt_clauses].lits = lits_tmp;

          i = 0;
          for (; *tmp != EOL; tmp++)
            {
              num = *tmp;
              clauses[cnt_clauses].lits[i++] = num;

              if (num < 0)
                vars[-num].neg_occs[vars[-num].num_neg_occs++] = cnt_clauses;
              else
                vars[num].pos_occs[vars[num].num_pos_occs++] = cnt_clauses;
            }
          assert (i == clauses[cnt_clauses].num_lits);
          assert (*tmp == EOL);
          tmp += 1;

          lits_tmp += clauses[cnt_clauses].num_lits;

          if (is_clause_empty (cnt_clauses))
            result = STATE_UNSAT;

          if ((unit = is_clause_unit (cnt_clauses)))
            {
              if (unit < 0)
                {
                  unit = -unit;
                  assignment = BTYPE_FALSE;
                }
              else
                assignment = BTYPE_TRUE;

              /* do not push any doubles */
              if (vars[unit].assignment == BTYPE_UNDEF)
                assign (unit, assignment, DLEVEL_0);
            }
          cnt_clauses += 1;
        }
    }
  free (stack.bp);
  free (stack_free_vars.bp);

  for (i = 1; !options.no_pures && i <= num_vars; i++)
    {
      if (vars[i].assignment == BTYPE_UNDEF && ((vars[i].num_pos_occs_cur == 0
                                                 && vars[i].num_neg_occs_cur >
                                                 0)
                                                || (vars[i].
                                                    num_neg_occs_cur == 0
                                                    && vars[i].
                                                    num_pos_occs_cur > 0)))
        {
          assert (vars[i].mode == VMODE_UNDEF);

          vars[i].mode = VMODE_PURE;
          if (vars[i].type == QTYPE_EXISTS)
            assignment =
              vars[i].num_pos_occs_cur == 0 ? BTYPE_FALSE : BTYPE_TRUE;
          else
            assignment =
              vars[i].num_neg_occs_cur == 0 ? BTYPE_FALSE : BTYPE_TRUE;

          assign (i, assignment, DLEVEL_0);
        }
    }

  return result;
}


/*
 * Mark all clauses previously unsatisfied as satisfied by var. Update 
 * occurence counters for all clauses satisfied by var. 
 */
void
mark_occs_sat (VarId var)
{
  assert (var > 0);
  assert (vars[var].assignment == BTYPE_TRUE ||
          vars[var].assignment == BTYPE_FALSE);

  int i, num_occs_tmp;
  ClauseId clause, *occs_tmp;

  /* check which occurences have to be updated (decremented) */
  if (vars[var].assignment == BTYPE_TRUE)
    {
      num_occs_tmp = vars[var].num_pos_occs;
      occs_tmp = vars[var].pos_occs;
    }
  else
    {
      num_occs_tmp = vars[var].num_neg_occs;
      occs_tmp = vars[var].neg_occs;
    }

  /* mark occurences as sat  */
  for (i = 0; i < num_occs_tmp; i++)
    {
      clause = occs_tmp[i];
      if (clauses[clause].sat_var == 0) /* clause yet unsatisfied */
        {
          clauses[clause].sat_var = var;        /* mark clause as sat by var */
          update_occs_cnt (clause, -1);
        }
    }
}


/* 
 * Update occurences of all literals in given clause by given offset.
 * Offset may be positive or negative.
 */
void
update_occs_cnt (ClauseId id, int offset)
{
  int i, phase;
  VarId var, sat_var;
  DLevel d_level;
  BType assignment;

  sat_var = clauses[id].sat_var;
  d_level = vars[sat_var].d_level;

  assert (sat_var > 0);

  for (i = 0; i < clauses[id].num_lits; i++)
    {
      phase = 1;
      var = clauses[id].lits[i];
      if (var < 0)
        {
          phase = 0;
          var = -var;
        }

      /* do not update occurence of satisfying variable */
      if (var == sat_var)
        continue;

      if (phase == 1)
        vars[var].num_pos_occs_cur += offset;
      else
        vars[var].num_neg_occs_cur += offset;

      assert (vars[var].num_pos_occs_cur >= 0);
      assert (vars[var].num_neg_occs_cur >= 0);
      assert (vars[var].num_pos_occs_cur <= vars[var].num_pos_occs);
      assert (vars[var].num_neg_occs_cur <= vars[var].num_neg_occs);

      /* do not check for pure literals during backtrack */
      if (offset < 0 && vars[var].assignment == BTYPE_UNDEF &&  /* don't push double */
          ((vars[var].num_pos_occs_cur == 0 && vars[var].num_neg_occs_cur > 0)
           ||
           (vars[var].num_neg_occs_cur == 0
            && vars[var].num_pos_occs_cur > 0)))
        {
          assert (d_level == DLEVEL_0 || d_level == DLEVEL_1);
          assert (vars[var].d_level == DLEVEL_UNDEF);
          assert (vars[var].mode == VMODE_UNDEF);

          vars[var].mode = VMODE_PURE;

          if (ABS_VAR_IS_EXISTS (var))
            {
              assignment =
                vars[var].num_pos_occs_cur == 0 ? BTYPE_FALSE : BTYPE_TRUE;
            }
          else
            {
              assignment =
                vars[var].num_neg_occs_cur == 0 ? BTYPE_FALSE : BTYPE_TRUE;
            }

          assign (var, assignment, d_level);
        }
    }
}


static int
is_clause_sat (ClauseId id)
{
  int i, phase;
  VarId var;

  if (clauses[id].is_tautology)
    return 1;

  for (i = 0; i < clauses[id].num_lits; i++)
    {
      phase = 1;
      var = clauses[id].lits[i];
      if (var < 0)
        {
          phase = 0;
          var = -var;
        }

      if ((vars[var].assignment == BTYPE_TRUE && phase == 1) ||
          (vars[var].assignment == BTYPE_FALSE && phase == 0))
        return 1;
    }

  return 0;
}


static int
is_clause_empty (ClauseId id)
{
  int i, phase;
  VarId var;
  QType type;

  if (clauses[id].is_tautology)
    return 0;

  for (i = 0; i < clauses[id].num_lits; i++)
    {
      phase = 1;
      var = clauses[id].lits[i];
      if (var < 0)
        {
          phase = 0;
          var = -var;
        }

      /* treat outer universal literals as existential if propagated variable
       * is existential and inner to these universals.  */
      type = vars[var].type;

      if (ABS_VAR_IS_EXISTS (var))
        type = QTYPE_EXISTS;

      /* no literal true and all existential literals false? */
      if ((vars[var].assignment == BTYPE_TRUE && phase == 1) ||
          (vars[var].assignment == BTYPE_FALSE && phase == 0) ||
          (type == QTYPE_EXISTS && vars[var].assignment == BTYPE_UNDEF))
        return 0;               /* not empty */
    }
  return 1;
}


static int
is_clause_unit (ClauseId id)
{
  int i, phase, lowest_level;
  Lit unit;
  VarId var, unit_id;
  QType type;

  if (clauses[id].is_tautology)
    return 0;

  lowest_level = num_scopes;
  unit = 0;

  /* check if clause is unit */
  for (i = 0; i < clauses[id].num_lits; i++)
    {
      phase = 1;
      var = clauses[id].lits[i];
      if (var < 0)
        {
          phase = 0;
          var = -var;
        }

      /* treat outer universal literals as existential if propagated variable
       * is existential and inner to these universals.  */
      type = vars[var].type;

      if (ABS_VAR_IS_EXISTS (var))
        type = QTYPE_EXISTS;

      /* no literal true? */
      if ((vars[var].assignment == BTYPE_TRUE && phase == 1) ||
          (vars[var].assignment == BTYPE_FALSE && phase == 0))
        return 0;

      /* all but one existential literal false? */
      if (vars[var].assignment == BTYPE_UNDEF)
        {
          assert (type == QTYPE_EXISTS || type == QTYPE_FORALL);
          if (type == QTYPE_EXISTS)
            {
              if (unit == 0)
                unit = clauses[id].lits[i];
              else
                return 0;
            }
          else
            {
              if (vars[var].s_level < lowest_level)
                lowest_level = vars[var].s_level;
            }
        }
    }

  unit_id = unit < 0 ? -unit : unit;
  /* either not even 1 existential literal is unassigned or there exists
   * at least one universal literal with lower scope level than the potential 
   * unit literal  */
  if (unit == 0 || lowest_level <= vars[unit_id].s_level)
    return 0;

  vars[unit_id].mode = VMODE_UNIT;
  assert (vars[unit_id].antecedent == INVALID_CLAUSE_ID);
  vars[unit_id].antecedent = id;

  return unit;
}


inline static void
assign (VarId var, BType assignment, DLevel d_level)
{
  assert (var > 0);
  assert (var <= num_vars);
  assert (assignment == BTYPE_TRUE || assignment == BTYPE_FALSE);
  assert (d_level == DLEVEL_0 || d_level == DLEVEL_1);
  assert (vars[var].type == QTYPE_EXISTS || vars[var].type == QTYPE_FORALL);
  assert (d_level == DLEVEL_1 || vars[var].mode == VMODE_UNIT ||
          vars[var].mode == VMODE_PURE);
  assert (vars[var].d_level == DLEVEL_UNDEF);
  assert (vars[var].assignment == BTYPE_UNDEF);

  vars[var].assignment = assignment;
  vars[var].d_level = d_level;
  assert (vars[var].trail_pos == INVALID_TRAIL_POS);
  vars[var].trail_pos = STACK_CNT (assigned_vars);

  PUSH (assigned_vars, var);

  if (vars[var].d_level == DLEVEL_0)
    {
      assigned_vars_l1 += 1;
      assert (assigned_vars_l1 == assigned_vars.sp);
    }

  assert (assigned_vars_l1 <= (assigned_vars.bp + assigned_vars.size));
}


static void
backtrack ()
{
  assert (assigned_vars.sp > assigned_vars_l1);

  int i, num_occs_tmp;
  VarId var = 0;
  ClauseId clause, *occs_tmp;

  while (assigned_vars.sp > assigned_vars_l1)
    {
      POP (assigned_vars, var);
      assert (vars[var].d_level != DLEVEL_UNDEF);
      assert (vars[var].d_level != DLEVEL_0);
      assert (vars[var].d_level == DLEVEL_1);

      if (vars[var].assignment == BTYPE_TRUE)
        {
          num_occs_tmp = vars[var].num_pos_occs;
          occs_tmp = vars[var].pos_occs;
        }
      else
        {
          num_occs_tmp = vars[var].num_neg_occs;
          occs_tmp = vars[var].neg_occs;
        }

      vars[var].d_level = DLEVEL_UNDEF;
      vars[var].assignment = BTYPE_UNDEF;
      vars[var].mode = VMODE_UNDEF;
      vars[var].is_propagated = 0;

      vars[var].antecedent = INVALID_CLAUSE_ID;
      assert (vars[var].trail_pos != INVALID_TRAIL_POS);
      vars[var].trail_pos = INVALID_TRAIL_POS;

      /* reset occurence counters */
      for (i = 0; !options.no_pures && i < num_occs_tmp; i++)
        {
          clause = occs_tmp[i];

          if (clauses[clause].sat_var == var)
            {
              update_occs_cnt (clause, 1);
              clauses[clause].sat_var = 0;
            }
        }
    }

  decision_var = 0;
  empty_clause = INVALID_CLAUSE_ID;
}


static QxBFState
propagate (VarId var)
{
  int i, num_occs_tmp;
  Lit unit;
  ClauseId clause, *occs_tmp;
  BType assignment;

  assert (var > 0);
  assert (vars[var].is_propagated == 0);
  assert (vars[var].assignment == BTYPE_TRUE ||
          vars[var].assignment == BTYPE_FALSE);
  assert (vars[var].d_level == DLEVEL_0 || vars[var].d_level == DLEVEL_1);
  assert (empty_clause == INVALID_CLAUSE_ID);

  assert (vars[var].d_level != DLEVEL_1
          || cur_bcp_propagations < cur_bcp_prop_bound);

  /* Ignore top-level propagations for bound and statistics. */
  if (vars[var].d_level == DLEVEL_1)
    {
      total_bcp_propagations++;
      cur_bcp_propagations++;
    }

  if (options.verbosity >= 2)
    fprintf (stderr, "propagate: %c%d mode %c level %d\n",
             vars[var].assignment == BTYPE_FALSE ? '-' : '+', var,
             vars[var].mode == VMODE_UNIT ? 'u' : vars[var].mode ==
             VMODE_PURE ? 'p' : 'd', vars[var].d_level);

  vars[var].is_propagated = 1;

  if (!options.no_pures)
    mark_occs_sat (var);

  if (vars[var].assignment == BTYPE_TRUE)
    {
      num_occs_tmp = vars[var].num_neg_occs;
      occs_tmp = vars[var].neg_occs;
    }
  else
    {
      num_occs_tmp = vars[var].num_pos_occs;
      occs_tmp = vars[var].pos_occs;
    }

  for (i = 0; i < num_occs_tmp; i++)
    {
      clause = occs_tmp[i];

      if (is_clause_sat (clause))
        continue;

      if (is_clause_empty (clause))
        {
          assert (empty_clause == INVALID_CLAUSE_ID);
          empty_clause = clause;
          return STATE_CONFLICT;
        }

      if ((unit = is_clause_unit (clause)))
        {
          if (unit < 0)
            {
              assignment = BTYPE_FALSE;
              unit = -unit;
            }
          else
            assignment = BTYPE_TRUE;

          assert (vars[unit].mode == VMODE_UNIT);
          assign (unit, assignment, vars[var].d_level);
        }
    }
  return STATE_UNDEF;
}


/* All scopes used for decisions, they are still traversed only once.
 * Note, that currently only one phase (true) is tried. */
static int
time_limit_reached (unsigned int *limit_ptr, double *total_time_ptr)
{
  return (*limit_ptr && *limit_ptr <= *total_time_ptr);
}


static void
sum_up_cur_time (double *cur_time_stamp, double *total_time)
{
  double cur_time = time_stamp ();
  *total_time = *total_time + (cur_time - *cur_time_stamp);
  *cur_time_stamp = cur_time;
}


static int
decide (void)
{
  assert (empty_clause == INVALID_CLAUSE_ID);
  VarId var;

  /* Maybe too costly to take time stamp at EVERY decision, so do this
     every N decisions. */
  if ((total_bcp_decisions & ((unsigned long long) (1 << 7) - 1)) == 0)
    {
      sum_up_cur_time (&cur_bcp_time_stamp, &total_bcp_time);
      /* Check if BCP time limit exceeded. */
      if (time_limit_reached (&options.bcp_time_limit, &total_bcp_time))
        return 0;
    }

  if (max_cur_bcp_propagations < cur_bcp_propagations)
    max_cur_bcp_propagations = cur_bcp_propagations;
  cur_bcp_propagations = 0;

  /* all scopes used for decisions, they are still traversed only once */
  for (;;)
    {
      if (scope_var_index >= scopes[scope_index].num_vars)
        {
          scope_index += 1;
          scope_var_index = 0;
          if (scope_index >= num_scopes)
            return 0;
        }

      var = scopes[scope_index].vars[scope_var_index++];
      if (!vars[var].is_propagated)
        break;
    }

  total_bcp_decisions++;

  if (options.verbosity >= 1)
    if ((total_bcp_decisions & ((unsigned long long) (1 << 14) - 1)) == 0)
      fprintf (stderr, "BCP-Decisions: %llu\n", total_bcp_decisions);

  assert (vars[var].assignment == BTYPE_UNDEF);
  assert (vars[var].d_level == DLEVEL_UNDEF);

  if (!tried)
    {
      assign (var, BTYPE_TRUE, DLEVEL_1);
      /* Want to try that variable again. */
      scope_var_index--;
      tried = 1;
    }
  else
    {
      assign (var, BTYPE_FALSE, DLEVEL_1);
      tried = 0;
    }

  /* Treat outer universal variables as existential ones. */
  if (!options.no_abs)
    decision_var = var;
  else
    decision_var = 0;

  if (options.verbosity >= 2)
    fprintf (stderr, "decide: %d\n", var);

  return 1;
}


static void
print_lits (Lit * start, Lit * end)
{
  assert (start <= end);
  Lit *p;
  for (p = start; p < end; p++)
    {
      fprintf (stderr, "%d ", *p);
    }
  fprintf (stderr, "0\n");
}


static void
print_assignment ()
{
  int *p;
  for (p = assigned_vars.bp; p < assigned_vars.sp; p++)
    {
      VarId var = *p;
      if (var < 0)
        var = -var;
      assert (var > 0);
      assert (vars[var].type == QTYPE_EXISTS
              || vars[var].type == QTYPE_FORALL);
      fprintf (stderr, "top-assigned: %c%d type %c mode %c level %d\n",
               vars[var].assignment == BTYPE_FALSE ? '-' : '+', var,
               vars[var].type == QTYPE_EXISTS ? 'e' : 'a',
               vars[var].mode == VMODE_UNIT ? 'u' : vars[var].mode ==
               VMODE_PURE ? 'p' : 'd', vars[var].d_level);
    }
}


static int
print_sorted_toplevel_cmp (const void *elem1, const void *elem2)
{
  int e1 = *((int *) elem1);
  int e2 = *((int *) elem2);
  if (e1 < e2)
    return -1;
  else if (e1 > e2)
    return 1;
  else
    return 0;
}


static void
print_sorted_toplevel ()
{
  const int topsize = STACK_CNT (assigned_vars);
  int toplevel[topsize];
  memset (toplevel, 0, topsize * sizeof (int));

  int *to = toplevel;
  int *p;
  for (p = assigned_vars.bp; p < assigned_vars.sp; p++)
    {
      assert (to < (toplevel + topsize));
      *to++ = *p;
    }

  qsort (toplevel, topsize, sizeof (int), print_sorted_toplevel_cmp);

  int *e;
  for (p = toplevel, e = p + topsize; p < e; p++)
    {
      VarId var = *p;
      if (var < 0)
        var = -var;
      assert (var > 0);
      assert (vars[var].type == QTYPE_EXISTS
              || vars[var].type == QTYPE_FORALL);
      fprintf (stderr, "top-assigned: %c%d type %c mode %c level %d\n",
               vars[var].assignment == BTYPE_FALSE ? '-' : '+', var,
               vars[var].type == QTYPE_EXISTS ? 'e' : 'a',
               vars[var].mode == VMODE_UNIT ? 'u' : vars[var].mode ==
               VMODE_PURE ? 'p' : 'd', vars[var].d_level);
    }
}


/* ---------- START: Q-RESOLUTION ---------- */

#ifndef NDEBUG
/* Return index of 'var' on assigned-stack. Used for assertion
   checking only. */
static unsigned int
get_var_trail_pos (VarId var)
{
  assert (var > 0);
  unsigned int pos = 0;
  Lit *p;
  for (p = assigned_vars.bp; p < assigned_vars.sp; p++)
    {
      VarId assigned_var = *p;
      if (assigned_var < 0)
        assigned_var = -assigned_var;
      if (assigned_var == var)
        return pos;
      pos++;
    }
  /* Variable 'var' not assigned. */
  return UINT_MAX;
}
#endif


static int
is_univ_lit_irreducible_in_clause (Lit lit, VarId ignore_var, Lit * start,
                                   Lit * end)
{
  assert (start <= end);
  assert (ignore_var >= 0);
  VarId var1 = lit;
  if (var1 < 0)
    var1 = -var1;
  assert (var1 > 0);
  assert (!ABS_VAR_IS_EXISTS (var1));

  /* Check if there is an existential literal which prevents
     literal 'var1' from being reduced by forall reduction. */
  Lit *p;
  for (p = start; p < end; p++)
    {
      VarId var2 = *p;
      if (var2 < 0)
        var2 = -var2;
      assert (var2 > 0);
      if (vars[var2].d_level == DLEVEL_0)
        {
          /* Ignore level-0 literals. */
          assert (vars[var2].assignment != BTYPE_UNDEF);
          assert (vars[var2].mode != VMODE_UNDEF);
          continue;
        }
      if (ABS_VAR_IS_EXISTS (var2) && (!ignore_var || ignore_var != var2)
          && vars[var1].s_level < vars[var2].s_level)
        {
          assert (p != end);
          break;
        }
    }
  return p != end;
}


#ifndef NDEBUG
static int
is_clause_forall_reduced (Lit * start, Lit * end)
{
  assert (start <= end);
  Lit *p1;
  for (p1 = start; p1 < end; p1++)
    {
      VarId var1 = *p1;
      if (var1 < 0)
        var1 = -var1;
      assert (var1 > 0);
      if (vars[var1].d_level == DLEVEL_0)
        {
          /* Ignore level-0 literals. */
          assert (vars[var1].assignment != BTYPE_UNDEF);
          assert (vars[var1].mode != VMODE_UNDEF);
          continue;
        }
      if (!ABS_VAR_IS_EXISTS (var1))
        {
          /* Check if there is an existential literal which prevents
             literal 'var1' from being reduced by forall reduction. */
          if (!(is_univ_lit_irreducible_in_clause (var1, 0, start, end)))
            return 0;
        }
    }
  return 1;
}


static int
has_clause_level0_lits (Lit * start, Lit * end)
{
  assert (start <= end);
  Lit *p;
  for (p = start; p < end; p++)
    {
      VarId var = *p;
      if (var < 0)
        var = -var;
      assert (var > 0);
      if (vars[var].d_level == DLEVEL_0)
        {
          assert (vars[var].assignment != BTYPE_UNDEF);
          assert (vars[var].mode != VMODE_UNDEF);
          return 1;
        }
    }
  return 0;
}
#endif

/* POSSIBLE OPTIMIZATION: can check for tautologies on-the-fly during
   collection of antecedent literals when marking all literals which are
   part of current working reason. */
static VarId
peek_resolvent_tautological (VarId pivot, Stack * qres_wreason,
                             Stack * qres_antecedent)
{
  /* Assumes that both working reason and antecedent are fully cleaned up. */
  assert (!has_clause_level0_lits (qres_wreason->bp, qres_wreason->sp));
  assert (is_clause_forall_reduced (qres_wreason->bp, qres_wreason->sp));
  assert (!has_clause_level0_lits (qres_antecedent->bp, qres_antecedent->sp));
  assert (is_clause_forall_reduced
          (qres_antecedent->bp, qres_antecedent->sp));
  assert (pivot > 0);

  Lit *p1;
  for (p1 = qres_wreason->bp; p1 < qres_wreason->sp; p1++)
    {
      VarId var1 = *p1;
      assert (var1);
      if (var1 < 0)
        var1 = -var1;
      /* Must ignore pivot from tautology check. */
      if (var1 == pivot)
        continue;
      Lit *p2;
      for (p2 = qres_antecedent->bp; p2 < qres_antecedent->sp; p2++)
        {
          /* Literals in working reason and antecedent are
             complementary. */
          if (*p1 == -*p2)
            return var1;
        }
    }
  return 0;
}

static int
lit_occurs_in_clause (Lit lit, Lit * start, Lit * end)
{
  assert (start <= end);
  Lit *p;
  for (p = start; p < end; p++)
    {
      if (*p == lit)
        return 1;
    }
  return 0;
}


#ifndef NDEBUG
/* Check if there are double or complementary literals in the
   clause represented by stack. */
static int
lits_unique_in_clause (Stack * stack)
{
  Lit *p1;
  for (p1 = stack->bp; p1 < stack->sp; p1++)
    {
      Lit *p2;
      for (p2 = p1 + 1; p2 < stack->sp; p2++)
        if (*p1 == *p2 || *p1 == -*p2)
          return 0;
    }
  return 1;
}
#endif


/* Resolve working reason with antecedent on pivot and store resolvent
   literals on stack. */
static void
resolve (VarId pivot, Stack * wreason, Stack * antecedent, Stack * resolvent)
{
  /* Assumes that both working reason and antecedent are fully cleaned up. */
  assert (!has_clause_level0_lits (wreason->bp, wreason->sp));
  assert (is_clause_forall_reduced (wreason->bp, wreason->sp));
  assert (!has_clause_level0_lits (antecedent->bp, antecedent->sp));
  assert (is_clause_forall_reduced (antecedent->bp, antecedent->sp));
  assert (pivot > 0);
  assert (ABS_VAR_IS_EXISTS (pivot));
  assert (STACK_CNT (*resolvent) == 0);

  /* Collect literals from working reason and add to resolvent. */
  assert (pivot > 0);
  Lit *p;
  for (p = wreason->bp; p < wreason->sp; p++)
    {
      VarId var = *p;
      if (var < 0)
        var = -var;
      assert (var > 0);
      if (var == pivot)
        continue;
      assert (ABS_VAR_IS_EXISTS (var) || vars[var].type == QTYPE_FORALL);
      /* Collect existential literals or universal literals which are
         irreducible in either the working reason or antecedent, where
         we ignore pivot variable for preventing reduction since pivot
         is not part of resolvent. */
      if (ABS_VAR_IS_EXISTS (var) ||
          (is_univ_lit_irreducible_in_clause (*p, pivot,
                                              wreason->bp, wreason->sp) ||
           is_univ_lit_irreducible_in_clause (*p, pivot,
                                              antecedent->bp,
                                              antecedent->sp)))
        PUSH ((*resolvent), *p);
    }

  /* Collect literals from antecedent and add to resolvent. Must
     prevent adding literals multiple times. */
  for (p = antecedent->bp; p < antecedent->sp; p++)
    {
      VarId var = *p;
      if (var < 0)
        var = -var;
      assert (var > 0);
      if (var == pivot)
        continue;
      assert (ABS_VAR_IS_EXISTS (var) || vars[var].type == QTYPE_FORALL);
      if (!(lit_occurs_in_clause (*p, resolvent->bp, resolvent->sp)) &&
          (ABS_VAR_IS_EXISTS (var) ||
           (is_univ_lit_irreducible_in_clause (*p, pivot,
                                               antecedent->bp, antecedent->sp)
            || is_univ_lit_irreducible_in_clause (*p, pivot, wreason->bp,
                                                  wreason->sp))))
        {
          PUSH ((*resolvent), *p);
        }
    }

  /* Resolvent does not have literals assigned at level 0, is fully
     forall-reduced (assertion checking outside this function), no double
     and complementary literals. */
  assert (lits_unique_in_clause (resolvent));
}


static void
collect_lits (Stack * stack, Lit * start, Lit * end, Lit ignore)
{
  assert (start <= end);
  VarId ignore_var = ignore;
  if (ignore_var < 0)
    ignore_var = -ignore_var;
  Lit *p;
  for (p = start; p < end; p++)
    {
      VarId var = *p;
      assert (var);
      if (var == ignore)
        continue;
      if (var < 0)
        var = -var;
      assert (var > 0);
      assert (ABS_VAR_IS_EXISTS (var) || vars[var].type == QTYPE_FORALL);
      /* Collect literals which are NOT assigned at level 0 and
         which are NOT reducible by forall reduction (the latter
         affects only universal literals). */
      if (vars[var].d_level != DLEVEL_0 &&
          (ABS_VAR_IS_EXISTS (var) ||
           is_univ_lit_irreducible_in_clause (var, ignore_var, start, end)))
        PUSH ((*stack), (*p));
    }
}


/* Stop resolution process if current resolvent either has exactly one
   literal or is empty. Check that the single literal is the inverted
   decision. */
static int
stop_res (Stack * working_reason)
{
  qres_derived_empty_clause =
    qres_derived_empty_clause || STACK_CNT (*working_reason) == 0;

  if (options.verbosity >= 2 && STACK_CNT (*working_reason) == 0)
    fprintf (stderr, "QRES derived empty clause.\n");

  /* NOTE: can happen that for decision x we get conflict, but qres
     does not produce -x first but another unit y. We could learn such
     y then later when deciding -y. In the following, we continue
     resolution until the inverted decision varaible is produced as a
     unit clause. */
  /* NOTE: if decision was universal, then can NEVER derive inverted
     decision as unit because would get empty clause by
     forall-reduction. */
  int derived_inverse_decision = 0;
  if (STACK_CNT (*working_reason) == 1)
    {
      Lit lit = *working_reason->bp;
      VarId var = lit;
      if (var < 0)
        var = -var;
      assert (var != 0);
      VarId dec_var = *assigned_vars_l1;
      if (dec_var < 0)
        dec_var = -dec_var;
      assert (vars[dec_var].mode != VMODE_UNIT &&
              vars[dec_var].mode != VMODE_PURE);
      if (var == dec_var)
        {
          assert (vars[dec_var].assignment == BTYPE_FALSE || lit < 0);
          assert (vars[dec_var].assignment == BTYPE_TRUE || lit > 0);
          derived_inverse_decision = 1;
          if (options.verbosity >= 2)
            fprintf (stderr, "QRES derived inverted decision-var.\n");
        }
    }

  return (STACK_CNT (*working_reason) == 0 || derived_inverse_decision);
}


static int
qres_aux (int (*stop_res_func) (Stack *), VarId ignore_pivot)
{
  assert (!ignore_pivot || ignore_pivot > 0);
  assert (!has_clause_level0_lits (qres_wreason.bp, qres_wreason.sp));
  assert (is_clause_forall_reduced (qres_wreason.bp, qres_wreason.sp));
  /* Initial working reason must not be empty and
     we must enter resolution-loop at least once. */
  assert (STACK_CNT (qres_wreason) > 0);
  assert (ignore_pivot || !stop_res_func (&qres_wreason));

  int result = 1;
  while (!stop_res_func (&qres_wreason))
    {
      assert (STACK_CNT (qres_wreason) > 0);
      assert (!has_clause_level0_lits (qres_wreason.bp, qres_wreason.sp));
      assert (is_clause_forall_reduced (qres_wreason.bp, qres_wreason.sp));

      if (options.verbosity >= 2)
        {
          fprintf (stderr, "wreason: ");
          print_lits (qres_wreason.bp, qres_wreason.sp);
        }

      /* Search for suitable pivot variable. */
      /* NOTE: we MUST search for pivots in reverse trail
         order. Otherwise, it might be impossible to generate a
         non-tautological resolvent. */
      ClauseId antecedent;
      unsigned int max_trail_pos = INVALID_TRAIL_POS;
      VarId max_trail_pivot = 0;
      Lit *pivot_pos;

      /* UPDATE: first search for maximum trail unit. If this produces
         a tautology, then try to get rid of literals which prevent
         forall-reduction of tautology literals. */
      for (pivot_pos = qres_wreason.bp; pivot_pos < qres_wreason.sp;
           pivot_pos++)
        {
          assert (!max_trail_pivot ||
                  (max_trail_pos != INVALID_TRAIL_POS &&
                   max_trail_pos < STACK_CNT (assigned_vars) &&
                   max_trail_pos == get_var_trail_pos (max_trail_pivot) &&
                   max_trail_pos == vars[max_trail_pivot].trail_pos));

          VarId pivot_candidate = *pivot_pos;
          if (pivot_candidate < 0)
            pivot_candidate = -pivot_candidate;
          assert (pivot_candidate > 0);
          assert (vars[pivot_candidate].trail_pos ==
                  get_var_trail_pos (pivot_candidate));

          if (pivot_candidate == ignore_pivot)
            continue;

          if (ABS_VAR_IS_EXISTS (pivot_candidate) &&
              vars[pivot_candidate].mode == VMODE_UNIT &&
              (!max_trail_pivot
               || max_trail_pos < vars[pivot_candidate].trail_pos))
            {
              max_trail_pivot = pivot_candidate;
              max_trail_pos = vars[pivot_candidate].trail_pos;
            }
        }

      if (!max_trail_pivot)
        {
          /* Did not find pivot variable, i.e. q-resolution neither
             was able to produce inverted decision literal as unit nor
             empty clause. */
          if (options.verbosity >= 2)
            fprintf (stderr, "QRES: did not find pivot!\n");
          result = 0;
          break;
        }

      /* Check if maximum trail unit produces tautology. */
      antecedent = vars[max_trail_pivot].antecedent;
      assert (antecedent != INVALID_CLAUSE_ID);
      assert (STACK_CNT (qres_antecedent) == 0);
      /* Collect antecedent literals. */
      collect_lits (&qres_antecedent, clauses[antecedent].lits,
                    clauses[antecedent].lits +
                    clauses[antecedent].num_lits, 0);
      assert (!has_clause_level0_lits
              (qres_antecedent.bp, qres_antecedent.sp));
      assert (is_clause_forall_reduced
              (qres_antecedent.bp, qres_antecedent.sp));

      VarId blocking = peek_resolvent_tautological
        (max_trail_pivot, &qres_wreason, &qres_antecedent);
      STACK_RESET (qres_antecedent);

      if (blocking)
        {
          /* Try to resolve on unit which prevents tautology literals from being forall-reduced. */
          max_trail_pos = INVALID_TRAIL_POS;
          max_trail_pivot = 0;
          for (pivot_pos = qres_wreason.bp; pivot_pos < qres_wreason.sp;
               pivot_pos++)
            {
              assert (!max_trail_pivot ||
                      (max_trail_pos != INVALID_TRAIL_POS &&
                       max_trail_pos < STACK_CNT (assigned_vars) &&
                       max_trail_pos == get_var_trail_pos (max_trail_pivot) &&
                       max_trail_pos == vars[max_trail_pivot].trail_pos));

              VarId pivot_candidate = *pivot_pos;
              if (pivot_candidate < 0)
                pivot_candidate = -pivot_candidate;
              assert (pivot_candidate > 0);
              assert (vars[pivot_candidate].trail_pos ==
                      get_var_trail_pos (pivot_candidate));

              if (pivot_candidate == ignore_pivot)
                continue;

              if (ABS_VAR_IS_EXISTS (pivot_candidate) &&
                  vars[pivot_candidate].mode == VMODE_UNIT &&
                  vars[blocking].s_level < vars[pivot_candidate].s_level &&
                  (!max_trail_pivot
                   || max_trail_pos < vars[pivot_candidate].trail_pos))
                {
                  /* Found an existential unit which is pivot candidate:
                     either we have not found a max-pivot so far or
                     current candidate is larger wrt. trail order. Must
                     check if resolvent of current working reason and
                     antecedent of that unit would produce a tautology. */
                  antecedent = vars[pivot_candidate].antecedent;
                  assert (antecedent != INVALID_CLAUSE_ID);
                  assert (STACK_CNT (qres_antecedent) == 0);

                  /* Collect antecedent literals. */
                  collect_lits (&qres_antecedent, clauses[antecedent].lits,
                                clauses[antecedent].lits +
                                clauses[antecedent].num_lits, 0);
                  assert (!has_clause_level0_lits
                          (qres_antecedent.bp, qres_antecedent.sp));
                  assert (is_clause_forall_reduced
                          (qres_antecedent.bp, qres_antecedent.sp));

                  if (!peek_resolvent_tautological
                      (pivot_candidate, &qres_wreason, &qres_antecedent))
                    {
                      max_trail_pivot = pivot_candidate;
                      max_trail_pos = vars[pivot_candidate].trail_pos;
                    }
                  STACK_RESET (qres_antecedent);
                }
            }
        }

      if (!max_trail_pivot)
        {
          /* Did not find pivot variable, i.e. q-resolution neither
             was able to produce inverted decision literal as unit nor
             empty clause. */
          if (options.verbosity >= 2)
            fprintf (stderr, "QRES: did not find pivot!\n");
          result = 0;
          break;
        }

      assert (max_trail_pos != INVALID_TRAIL_POS &&
              max_trail_pos < STACK_CNT (assigned_vars) &&
              max_trail_pos == get_var_trail_pos (max_trail_pivot) &&
              max_trail_pos == vars[max_trail_pivot].trail_pos);

      /* Collect antecedent literals. */
      antecedent = vars[max_trail_pivot].antecedent;
      assert (antecedent != INVALID_CLAUSE_ID);
      assert (STACK_CNT (qres_antecedent) == 0);
      collect_lits (&qres_antecedent, clauses[antecedent].lits,
                    clauses[antecedent].lits + clauses[antecedent].num_lits,
                    0);
      assert (!has_clause_level0_lits
              (qres_antecedent.bp, qres_antecedent.sp));
      assert (is_clause_forall_reduced
              (qres_antecedent.bp, qres_antecedent.sp));

      if (options.verbosity >= 2)
        {
          fprintf (stderr, "pivot: %d\n", max_trail_pivot);
          fprintf (stderr, "antecedent: ");
          print_lits (qres_antecedent.bp, qres_antecedent.sp);
          fprintf (stderr, "\n");
        }

      /* Stack 'qres_tmp' will hold resolvent's literals. */
      resolve (max_trail_pivot, &qres_wreason, &qres_antecedent, &qres_tmp);

      STACK_RESET (qres_wreason);
      STACK_RESET (qres_antecedent);

      /* Swap stacks. In next iteration, again 'qres_wreason' will
         hold literals of working reason. */
      Stack tmp = qres_tmp;
      qres_tmp = qres_wreason;
      qres_wreason = tmp;
    }

  return result;
}


/* When encountering an empty clause after a decision 'x', check if
   unit clause '-x' can be derived by q-resolution. If so, we may keep
   '-x' as learnt unit clause, otherwise we must NOT learn '-x'. The
   idea is to carry out BCP on the original formula but use
   q-resolution to check if conflicts are valid. */
static int
is_unit_valid ()
{
  assert (empty_clause != INVALID_CLAUSE_ID);
  assert (is_clause_empty (empty_clause));
  assert (STACK_CNT (qres_wreason) == 0);
  assert (STACK_CNT (qres_antecedent) == 0);
  assert (STACK_CNT (qres_tmp) == 0);
  int result = 1;

  if (options.verbosity >= 2)
    {
      fprintf (stderr, "\nFunction '%s':\n", __func__);
      fprintf (stderr, "\nbefore qres:\n");
      print_assignment ();
      fprintf (stderr, "\n");
      fprintf (stderr, "empty clause: ");
      print_lits (clauses[empty_clause].lits,
                  clauses[empty_clause].lits +
                  clauses[empty_clause].num_lits);
      fprintf (stderr, "\n");
    }

  /* Collect literals from empty clause as current working reason. */
  collect_lits (&qres_wreason, clauses[empty_clause].lits,
                clauses[empty_clause].lits + clauses[empty_clause].num_lits,
                0);

  result = qres_aux (stop_res, 0);

  STACK_RESET (qres_wreason);
  assert (STACK_CNT (qres_wreason) == 0);
  assert (STACK_CNT (qres_antecedent) == 0);
  assert (STACK_CNT (qres_tmp) == 0);

  return result;
}

/* ---------- END: Q-RESOLUTION ---------- */


static QxBFState
level0_bcp ()
{
  QxBFState state;
  VarId var;
  /* decision level 0 */
  for (; next_to_propagate < assigned_vars_l1; next_to_propagate++)
    {
      var = *next_to_propagate;
      assert (var > 0);
      assert (var <= num_vars);
      assert (!vars[var].is_propagated);
      assert (vars[var].mode == VMODE_UNIT || vars[var].mode == VMODE_PURE);
      assert (empty_clause == INVALID_CLAUSE_ID);

#ifndef NDEBUG
      assert_check_pure_literals (__LINE__);
#endif

      state = propagate (var);

#ifndef NDEBUG
      assert_check_pure_literals (__LINE__);
#endif

      assert (vars[var].is_propagated);

      if (state == STATE_CONFLICT)      /* done */
        return STATE_UNSAT;
    }
  return STATE_UNDEF;
}


static QxBFState
bcp (void)
{
  VarId var = 0;
  QxBFState state = STATE_UNDEF;
  BType assignment = BTYPE_UNDEF;
  unsigned int old_learnt_from_failed_lits = learnt_from_failed_lits;

  /* Reset indices used in 'decide'-function. */
  scope_var_index = 0;
  scope_index = 0;

#ifndef NDEBUG
  assert_check_pure_literals (__LINE__);
#endif

  for (;;)
    {
      /* decision level 0 */
      state = level0_bcp ();

      if (state == STATE_UNSAT)
        break;

      assert (next_to_propagate == assigned_vars_l1);

#ifndef NDEBUG
      assert_check_is_propagated ();
#endif

      if (!decide ())           /* all vars possible have been used as decision */
        break;

#ifndef NDEBUG
      assert_check_pure_literals (__LINE__);
#endif

      QXBF_ABORT ((cur_bcp_propagations != 0),
                  "unexpected number of propagations!");
      assert (0 < cur_bcp_prop_bound
              && cur_bcp_prop_bound <= max_bcp_prop_bound);
      assert (0 < max_bcp_prop_bound
              && max_bcp_prop_bound <= (unsigned int) num_vars);

      for (;
           next_to_propagate < assigned_vars.sp
           && cur_bcp_propagations < cur_bcp_prop_bound; next_to_propagate++)
        {
          var = *next_to_propagate;

          assert (var > 0);
          assert (var <= num_vars);
          assert (!vars[var].is_propagated);
          assert (empty_clause == INVALID_CLAUSE_ID);

#ifndef NDEBUG
          assert_check_pure_literals (__LINE__);
#endif

          state = propagate (var);

#ifndef NDEBUG
          assert_check_pure_literals (__LINE__);
#endif

          /* Value of 'assignment' will be used to indicate whether a
             conflict is valid or not. This is relevant only if we do not
             treat outer universal variables as existential ones. */
          assert (assignment == BTYPE_UNDEF);

          if (state == STATE_CONFLICT)
            {
              if (!options.no_abs)
                {
                  /* Abstraction-based FL is sound by construction. */
                  assert (is_unit_valid ());
                  var = *assigned_vars_l1;
                  assignment = vars[var].assignment;
                  assert (assignment != BTYPE_UNDEF);
                }
              else
                {
                  /* When NOT using abstraction, we have to check if
                     inverted decision can be produced by
                     Q-resolution. */
                  if (is_unit_valid ())
                    {
                      var = *assigned_vars_l1;
                      assignment = vars[var].assignment;
                      assert (assignment != BTYPE_UNDEF);
                    }
                }
              /* Must backtrack anyway. */
              break;
            }
        }

#ifndef NDEBUG
      assert_check_pure_literals (__LINE__);
#endif

      backtrack ();

#ifndef NDEBUG
      assert_check_pure_literals (__LINE__);
#endif

      next_to_propagate = assigned_vars.sp;

      if (state == STATE_CONFLICT && assignment != BTYPE_UNDEF)
        {
          /* Detected a valid conflict. */
          learnt_from_failed_lits++;
          tried = 0;
          assert (vars[var].assignment == BTYPE_UNDEF);

          if (vars[var].type != QTYPE_FORALL)
            {
              assignment =
                (assignment == BTYPE_TRUE) ? BTYPE_FALSE : BTYPE_TRUE;
              vars[var].mode = VMODE_UNIT;
              assign (var, assignment, DLEVEL_0);
              if (options.verbosity >= 2)
                fprintf (stderr, "BCP (round %d): pushed learnt unit %c%d\n",
                         bcp_rounds, assignment == BTYPE_FALSE ? '-' : '+',
                         var);
            }
          else
            {
              /* Universal unit is not pushed, can return UNSAT immediately. */
              if (options.verbosity >= 2)
                fprintf (stderr, "BCP (round %d): universal unit %c%d\n",
                         bcp_rounds, assignment == BTYPE_FALSE ? '-' : '+',
                         var);
              state = STATE_UNSAT;
              break;
            }
          assignment = BTYPE_UNDEF;
#ifndef NDEBUG
          assert_check_pure_literals (__LINE__);
#endif
        }
    }

  if (options.verbosity >= 1)
    {
      fprintf (stderr, "BCP: %d failed-lit units in round %d\n",
               learnt_from_failed_lits - old_learnt_from_failed_lits,
               bcp_rounds);
      if (time_limit_reached (&options.bcp_time_limit, &total_bcp_time))
        fprintf (stderr, "BCP: time limit %d <= %f current reached\n",
                 options.bcp_time_limit, total_bcp_time);
    }

  return state;
}


/* Append top-level units only, do not simplify formula. */
static void
print_result_append (FILE * out, QxBFState state)
{
  int i, j, num_units, *tmp;
  char qtype;
  assert (!cnf_unsat || state == STATE_UNSAT);

  if (state == STATE_UNSAT)
    {
      /* Treat UNSAT-result separately. */
      fprintf (out, "p cnf 1 2\n");
      fprintf (out, "e 1 0\n");
      fprintf (out, "-1 0\n");
      fprintf (out, "1 0\n");
      return;
    }

  assert (next_to_propagate == assigned_vars_l1);
  assert (next_to_propagate == assigned_vars.sp);

  num_units = assigned_vars_l1 - assigned_vars_l0;      /* units + pure lits */
  for (tmp = assigned_vars_l0; tmp < assigned_vars_l1; tmp++)
    {
      assert (vars[*tmp].assignment == BTYPE_TRUE ||
              vars[*tmp].assignment == BTYPE_FALSE);

      if (options.verbosity >= 1 && vars[*tmp].type == QTYPE_FORALL &&
          vars[*tmp].mode == VMODE_UNIT)
        fprintf (stderr, "Output:  top-level contains universal unit!\n");

      if (vars[*tmp].mode == VMODE_PURE)
        {
          assert ((vars[*tmp].num_pos_occs_cur == 0 &&
                   vars[*tmp].num_neg_occs_cur >= 0) ||
                  (vars[*tmp].num_pos_occs_cur >= 0 &&
                   vars[*tmp].num_neg_occs_cur == 0));
          num_units -= 1;
        }
    }
  assert (num_units >= 0);

  /* preamble */
  fprintf (out, "p cnf %d %d\n", num_vars, num_clauses + num_units);

  /* scopes */
  for (i = 0; i < num_scopes; i++)
    {
      switch (scopes[i].type)
        {
        case QTYPE_EXISTS:
          qtype = QDIMACS_EXISTS;
          break;
        case QTYPE_FORALL:
          qtype = QDIMACS_FORALL;
          break;
        default:
          qtype = ' ';
          break;
        }
      fprintf (out, "%c ", qtype);

      for (j = 0; j < scopes[i].num_vars; j++)
        fprintf (out, "%d ", scopes[i].vars[j]);
      fprintf (out, "0\n");
    }

  /* units found */
  for (tmp = assigned_vars_l0; tmp < assigned_vars_l1; tmp++)
    {
      if (vars[*tmp].mode == VMODE_UNIT)
        {
          if (vars[*tmp].assignment == BTYPE_FALSE)
            fprintf (out, "-%d %c\n", *tmp, QDIMACS_EOL);
          else
            fprintf (out, "%d %c\n", *tmp, QDIMACS_EOL);
        }
    }

  /* clauses */
  for (i = 0; i < num_clauses; i++)
    {
      for (j = 0; j < clauses[i].num_lits; j++)
        fprintf (out, "%d ", clauses[i].lits[j]);
      fprintf (out, "%c\n", QDIMACS_EOL);
    }
}


/* Print cleaned-up formula, i.e. skip clauses and literals disabled
   by top-level assignment. Assumes that BCP has been run until
   saturation, unless formula was found UNSAT during BCP. */
static void
print_result_full_clean (FILE * out, QxBFState state)
{
  int i, j, phase, num_sat_clauses, num_occs, *tmp, *occs_tmp;
  char qtype;
  VarId var;
  assert (!cnf_unsat || state == STATE_UNSAT);

  assert (next_to_propagate == assigned_vars_l1);
  assert (next_to_propagate == assigned_vars.sp);

  /* find clauses sat by top-level */
  num_sat_clauses = 0;
  for (tmp = assigned_vars_l0; tmp < assigned_vars_l1; tmp++)
    {
      assert (vars[*tmp].assignment == BTYPE_TRUE ||
              vars[*tmp].assignment == BTYPE_FALSE);

      if (vars[*tmp].type == QTYPE_FORALL && vars[*tmp].mode == VMODE_UNIT)
        {
          if (options.verbosity >= 1)
            fprintf (stderr, "Output:  top-level contains universal unit!\n");
          state = STATE_UNSAT;
          break;
        }

      if (vars[*tmp].assignment == BTYPE_TRUE)
        {
          occs_tmp = vars[*tmp].pos_occs;
          num_occs = vars[*tmp].num_pos_occs;
        }
      else
        {
          occs_tmp = vars[*tmp].neg_occs;
          num_occs = vars[*tmp].num_neg_occs;
        }

      for (j = 0; j < num_occs; j++)
        {
          if (!clauses[occs_tmp[j]].is_pure_sat)
            {
              num_sat_clauses += 1;
              clauses[occs_tmp[j]].is_pure_sat = 1;
            }
        }
    }

  if (state == STATE_UNSAT)
    {
      /* Treat UNSAT-result separately. */
      fprintf (out, "p cnf 1 2\n");
      fprintf (out, "e 1 0\n");
      fprintf (out, "-1 0\n");
      fprintf (out, "1 0\n");
      return;
    }

  assert (num_clauses - num_sat_clauses >= 0);
  /* preamble */
  fprintf (out, "p cnf %d %d\n", num_vars, num_clauses - num_sat_clauses);

  //TODO? check for empty scopes (all vars assigned)

  /* scopes */
  for (i = 0; i < num_scopes; i++)
    {
      switch (scopes[i].type)
        {
        case QTYPE_EXISTS:
          qtype = QDIMACS_EXISTS;
          break;
        case QTYPE_FORALL:
          qtype = QDIMACS_FORALL;
          break;
        default:
          qtype = ' ';
          break;
        }
      fprintf (out, "%c ", qtype);

      for (j = 0; j < scopes[i].num_vars; j++)
        fprintf (out, "%d ", scopes[i].vars[j]);
      fprintf (out, "0\n");
    }

  /* clauses */
  for (i = 0; i < num_clauses; i++)
    {
      if (!clauses[i].is_pure_sat)
        {
          for (j = 0; j < clauses[i].num_lits; j++)
            {
              phase = 1;
              var = clauses[i].lits[j];
              if (var < 0)
                {
                  phase = 0;
                  var = -var;
                }

              assert (!(phase == 0 && vars[var].assignment == BTYPE_FALSE));
              assert (!(phase == 1 && vars[var].assignment == BTYPE_TRUE));
              assert (vars[var].mode != VMODE_UNDEF
                      || vars[var].assignment == BTYPE_UNDEF);
              assert (vars[var].assignment != BTYPE_UNDEF
                      || vars[var].mode == VMODE_UNDEF);

              /* Print only unassigned literals. */
              if (vars[var].mode == VMODE_UNDEF)
                fprintf (out, "%d ", clauses[i].lits[j]);
            }
          /* Must have printed at least one literal, otherwise would have
             detected conflict. */
          assert (j > 0);
          fprintf (out, "%c\n", QDIMACS_EOL);
        }
    }
}


static void
print_result_clean (FILE * out, QxBFState state)
{
  int i, j, phase, num_units, num_sat_clauses, num_occs, *tmp, *occs_tmp;
  char qtype;
  VarId var;
  assert (!cnf_unsat || state == STATE_UNSAT);

  if (state == STATE_UNSAT)
    {
      /* Treat UNSAT-result separately. */
      fprintf (out, "p cnf 1 2\n");
      fprintf (out, "e 1 0\n");
      fprintf (out, "-1 0\n");
      fprintf (out, "1 0\n");
      return;
    }

  assert (next_to_propagate == assigned_vars_l1);
  assert (next_to_propagate == assigned_vars.sp);

  /* find clauses sat by pure literal */
  num_units = assigned_vars_l1 - assigned_vars_l0;      /* units + pure lits */
  num_sat_clauses = 0;
  for (tmp = assigned_vars_l0; tmp < assigned_vars_l1; tmp++)
    {
      assert (vars[*tmp].assignment == BTYPE_TRUE ||
              vars[*tmp].assignment == BTYPE_FALSE);

      if (vars[*tmp].mode == VMODE_PURE)
        {
          assert ((vars[*tmp].num_pos_occs_cur == 0 &&
                   vars[*tmp].num_neg_occs_cur >= 0) ||
                  (vars[*tmp].num_pos_occs_cur >= 0 &&
                   vars[*tmp].num_neg_occs_cur == 0));
          num_units -= 1;

          if (vars[*tmp].assignment == BTYPE_TRUE)
            {
              occs_tmp = vars[*tmp].pos_occs;
              num_occs = vars[*tmp].num_pos_occs;
            }
          else
            {
              occs_tmp = vars[*tmp].neg_occs;
              num_occs = vars[*tmp].num_neg_occs;
            }

          for (j = 0; j < num_occs; j++)
            {
              if (!clauses[occs_tmp[j]].is_pure_sat)
                {
                  num_sat_clauses += 1;
                  clauses[occs_tmp[j]].is_pure_sat = 1;
                }
            }
        }
    }

  assert (num_clauses + num_units - num_sat_clauses >= 0);
  /* preamble */
  fprintf (out, "p cnf %d %d\n",
           num_vars, num_clauses + num_units - num_sat_clauses);

  /* scopes */
  for (i = 0; i < num_scopes; i++)
    {
      switch (scopes[i].type)
        {
        case QTYPE_EXISTS:
          qtype = QDIMACS_EXISTS;
          break;
        case QTYPE_FORALL:
          qtype = QDIMACS_FORALL;
          break;
        default:
          qtype = ' ';
          break;
        }
      fprintf (out, "%c ", qtype);

      for (j = 0; j < scopes[i].num_vars; j++)
        fprintf (out, "%d ", scopes[i].vars[j]);
      fprintf (out, "0\n");
    }

  /* units found */
  for (tmp = assigned_vars_l0; tmp < assigned_vars_l1; tmp++)
    {
      if (vars[*tmp].mode == VMODE_UNIT)
        {
          if (vars[*tmp].assignment == BTYPE_FALSE)
            fprintf (out, "-%d %c\n", *tmp, QDIMACS_EOL);
          else
            fprintf (out, "%d %c\n", *tmp, QDIMACS_EOL);
        }
    }

  /* clauses */
  for (i = 0; i < num_clauses; i++)
    {
      if (!clauses[i].is_pure_sat)
        {
          for (j = 0; j < clauses[i].num_lits; j++)
            {
              phase = 1;
              var = clauses[i].lits[j];
              if (var < 0)
                {
                  phase = 0;
                  var = -var;
                }

              assert (vars[var].mode != VMODE_PURE ||
                      !(phase == 0 && vars[var].assignment == BTYPE_FALSE));
              assert (vars[var].mode != VMODE_PURE ||
                      !(phase == 1 && vars[var].assignment == BTYPE_TRUE));

              if (vars[var].mode != VMODE_PURE)
                fprintf (out, "%d ", clauses[i].lits[j]);
            }
          fprintf (out, "%c\n", QDIMACS_EOL);
        }
    }
}


static void
cleanup (void)
{
  if (vars != NULL)
    free (vars);
  if (clauses != NULL)
    free (clauses);
  if (literals != NULL)
    free (literals);
  if (scopes != NULL)
    free (scopes);
  if (scope_vars != NULL)
    free (scope_vars);
  if (occurences != NULL)
    free (occurences);
  if (assigned_vars.bp != NULL)
    free (assigned_vars.bp);
  if (qres_wreason.bp != NULL)
    free (qres_wreason.bp);
  if (qres_antecedent.bp != NULL)
    free (qres_antecedent.bp);
  if (qres_tmp.bp != NULL)
    free (qres_tmp.bp);
}


void
print_usage (void)
{
  printf ("qxbf [<options>] [<file>]\n\n");
  printf ("  file:\n");
  printf ("    a QBF in QDIMACS format (stdin if not given)\n\n");
  printf ("  options:\n");
  printf
    ("    -h, --help  			print this message and exit\n");
  printf ("    --version   			print version\n");
  printf
    ("    -s <num>     			(INACTIVE OPTION) set random seed (default: 0)\n");
  printf
    ("    -v          			increase verbosity incrementally (default: 0)\n");
  printf
    ("    --no-pures  			disable pure literal detection\n");
  printf
    ("    --no-abs    			disable abstraction-based failed literal detection\n");
  printf
    ("    --no-bcp    			disable BCP-based failed literal detection\n");
  printf
    ("    --no-sat				disable SAT-based failed literal detection\n");
  printf
    ("    --print-toplevel			print detected fixed assignments\n");

  printf
    ("    --init-bcp-prop-limit=<num>		initial BCP propagation limit (default: 0)\n");
  printf
    ("    --bcp-prop-limit=<num>		max. BCP propagation limit (default: 4096)\n");

  printf
    ("    --init-sat-dec-limit=<num>		initial SAT solver decision limit (default: 1000)\n");
  printf
    ("    --sat-dec-limit=<num>		max. SAT solver decision limit (default: 1000)\n");
  printf
    ("    --sat-ignore-dist=<num>		periodically mark literals to be ignored after\n");
  printf
    ("					 2^{<num>} CNF models were found (default: 1)\n");
  printf
    ("    --sat-sweep-calls=<num>		<num> initial SAT solver calls to mark\n");
  printf
    ("					 literals to be ignored (default: 1)\n");
  printf
    ("    --sat-sweep-dec-limit=<num>		allow at most <num> decisions in initial SAT\n");
  printf
    ("					 solver calls (default: 500000)\n");
  printf
    ("    --sat-dec-limit-inc=<num>		allow <num> SAT solver decisions more\n");
  printf
    ("					 for SAT-based FL than determined in initial SAT\n");
  printf
    ("					 solver calls (default: 8000)\n");
  printf
    ("    --sat-time-threshold=<num>		use static decision limit if initial SAT solver calls\n");
  printf
    ("					 take longer than <num>/1000 sec. (default: num == 1000)\n");
  printf
    ("    --sat-time-limit=<num>		max. <num> seconds for SAT-based FL (default: 40)\n");
  printf
    ("    --bcp-time-limit=<num>		max. <num> seconds for BCP-based FL (default: 40)\n");
  printf ("\n");
}


void
print_version (void)
{
  fprintf (stderr, "%s\n", VERSION);
}


/* Export internal top-level assignment to SAT solver. */
static void
export_qxbf_toplevel (Lit * start, Lit * end)
{
  Lit *p, *e;
  for (p = start, e = end; p < e; p++)
    {
      VarId var = *p;
      assert (var > 0);
      assert (vars[var].d_level == DLEVEL_0);
      assert (vars[var].assignment != BTYPE_UNDEF);
      if (!picosat_deref_toplevel (var))
        {
          Lit lit = vars[var].assignment == BTYPE_FALSE ? -var : var;
          picosat_add (lit);
          picosat_add (0);
        }
    }
}


/* Export plain CNF to SAT solver. */
static void
sat_based_probing_setup ()
{
  int i;
  for (i = 0; i < num_clauses; i++)
    {
      int j;
      for (j = 0; j < clauses[i].num_lits; j++)
        picosat_add (clauses[i].lits[j]);
      picosat_add (0);
    }
  export_qxbf_toplevel (assigned_vars.bp, assigned_vars_l1);
}


/* Literals which are part of a SAT model of the CNF can be skipped
   when testing for failed literals. */
static void
mark_ignore_vars_by_model (int force_print_stats)
{
  if (picosat_res () == PICOSAT_SATISFIABLE)
    {
      unsigned int cnt = 0;
      int i;
      for (i = 1; i < num_vars; i++)
        {
          int val = picosat_deref (i);
          if (val)
            {
              assert (val == -1 || val == 1);
              if (val == 1)
                {
                  if (!vars[i].sat_test_pos_ignore)
                    cnt++;
                  vars[i].sat_test_pos_ignore = 1;
                }
              else
                {
                  if (!vars[i].sat_test_neg_ignore)
                    cnt++;
                  vars[i].sat_test_neg_ignore = 1;
                }
            }
        }
      total_ignore_literals += cnt;
      if ((force_print_stats && options.verbosity >= 1)
          || options.verbosity >= 2)
        fprintf (stderr,
                 "SAT-Tests ignore-literals - new: %d, total: %d (%f)\n", cnt,
                 total_ignore_literals,
                 num_vars ? total_ignore_literals / (float) (num_vars *
                                                             2) : 0);
    }
}


static int
sat_based_probing_assume (Lit assumption)
{
  total_sat_tests++;

  if ((total_sat_tests & ((unsigned long long) (1 << 10) - 1)) == 0)
    if (options.verbosity >= 1)
      fprintf (stderr, "SAT-Tests: %llu ignore-ratio %f\n",
               total_sat_tests,
               num_vars ? total_ignore_literals / (float) (num_vars * 2) : 0);

  /* Add assumption and run SAT test. */
  picosat_assume (assumption);
  int dec_limit = cur_sat_prop_bound ? (int) cur_sat_prop_bound : -1;
  double sat_start_time = 0;
  if (options.verbosity >= 2)
    sat_start_time = time_stamp ();
  int res = picosat_sat (dec_limit);

  /* Maybe too costly to take time stamp at EVERY test, so do this
     every N tests. */
  if ((total_sat_tests & ((unsigned long long) (1 << 7) - 1)) == 0)
    sum_up_cur_time (&cur_sat_time_stamp, &total_sat_time);

  double sat_time = 0;
  if (options.verbosity >= 2)
    sat_time = time_stamp () - sat_start_time;

  if (options.verbosity >= 2)
    fprintf (stderr, "SAT-based probing: assumed %d, result %d,"
             " picosat: time %f, propagations %llu\n\n",
             assumption, res, sat_time, picosat_propagations ());

  if (res == PICOSAT_SATISFIABLE)
    {
      total_sat_solver_models++;
      /* Mark variables which can be skipped. */
      if (options.sat_ignore_dist &&
          ((total_sat_solver_models &
            ((unsigned int) (1 << options.sat_ignore_dist) - 1)) == 0))
        mark_ignore_vars_by_model (0);
    }
  else if (res == PICOSAT_UNSATISFIABLE)
    {
      if (options.verbosity >= 2)
        fprintf (stderr, "SAT-based probing: UNSAT for assumption %d\n",
                 assumption);
      learnt_from_unsat_tests++;
      VarId var = assumption < 0 ? -assumption : assumption;
      assert (!vars[var].learnt_from_unsat_test);
      vars[var].learnt_from_unsat_test = 1;
    }
  else
    {
      assert (res == PICOSAT_UNKNOWN);
    }

  return res;
}


static QxBFState
check_failed_assumption (VarId assumption)
{
  /* Check for UNSAT in absence of assumption, can then immediately
     return UNSAT. */
  if (!picosat_failed_assumption (assumption))
    {
      cnf_unsat = 1;
      return STATE_UNSAT;
    }
  else
    {
      picosat_add (-assumption);
      picosat_add (0);
    }
  return STATE_UNDEF;
}


/* Collect top-level literals of Picosat and add to internal assignment. */
static void
import_sat_solver_toplevel ()
{
  int i;
  for (i = 1; i <= num_vars; i++)
    {
      int val;
      if (!vars[i].is_propagated && (val = picosat_deref_toplevel (i)))
        {
          if (!vars[i].learnt_from_unsat_test)
            learnt_from_sat_solver++;
          assert (val == 1 || val == -1);
          assert (vars[i].assignment == BTYPE_UNDEF);
          if (options.verbosity >= 1 && vars[i].type == QTYPE_FORALL)
            fprintf (stderr, "SAT: PicoSAT derived universal unit!\n");
          BType assignment = (val == 1) ? BTYPE_TRUE : BTYPE_FALSE;
          vars[i].mode = VMODE_UNIT;
          assign (i, assignment, DLEVEL_0);
          if (options.verbosity >= 2)
            fprintf (stderr, "SAT: pushed learnt unit %c%d\n",
                     assignment == BTYPE_FALSE ? '-' : '+', i);
        }
    }
}


static QxBFState
sat_based_probing_aux ()
{
  QxBFState state = STATE_UNDEF;
  int si;
  for (si = 0; si < num_scopes && state == STATE_UNDEF &&
       !time_limit_reached (&options.sat_time_limit, &total_sat_time); si++)
    {
      /* For all scopes. */
      int vi;
      for (vi = 0; vi < scopes[si].num_vars && state == STATE_UNDEF &&
           !time_limit_reached (&options.sat_time_limit, &total_sat_time);
           vi++)
        {
          /* For all variables in scope except ones which are already
             assigned either internally or in Picosat. */
          VarId var = scopes[si].vars[vi];
          if (!vars[var].is_propagated && !picosat_deref_toplevel (var) &&
              !(vars[var].sat_test_neg_ignore
                && vars[var].sat_test_pos_ignore))
            {
              VarId assumption;
              if (vars[var].sat_test_pos_ignore)
                assumption = -var;
              else
                assumption = var;
              int res = sat_based_probing_assume (assumption);
              if (res != PICOSAT_UNSATISFIABLE)
                {
                  if (!vars[var].sat_test_neg_ignore
                      && !vars[var].sat_test_pos_ignore)
                    {
                      assumption = -assumption;
                      res = sat_based_probing_assume (assumption);
                      if (res == PICOSAT_UNSATISFIABLE)
                        state = check_failed_assumption (assumption);
                    }
                }
              else
                state = check_failed_assumption (assumption);
            }
          assert (state == STATE_UNDEF || state == STATE_UNSAT);
        }
    }

  if (state == STATE_UNDEF)
    {
      /* NOTE: if state == UNSAT, then we do not collect
         PicoSAT's top-level assignment. Hence, only a trivial QBF will
         be written to output file, but not the top-level assignment of
         QxBF which possibly would justify the UNSAT-result. */
      assert (next_to_propagate == assigned_vars_l1);
      import_sat_solver_toplevel ();
      unsigned int old_size = STACK_CNT (assigned_vars);
      /* Propagate any collected literals internally. */
      state = level0_bcp ();
      assert (next_to_propagate == assigned_vars_l1);

      /* NOTE: if we export any assignmentd, then change of top-level
         occurred and we will enter another round. */
      if (state == STATE_UNDEF)
        export_qxbf_toplevel (assigned_vars.bp + old_size, assigned_vars_l1);
    }

  if (options.verbosity >= 1
      && time_limit_reached (&options.sat_time_limit, &total_sat_time))
    fprintf (stderr, "SAT: time limit %d <= %f current reached\n",
             options.sat_time_limit, total_sat_time);

  return state;
}


/* PicoSAT is called once to estimate cost of SAT solver call. */
static int
get_picosat_decisions ()
{
  int decisions = -1;
  char *scratch_filename = "tmp";
  int pid = getpid ();
  char buffer[1024] = { 0 };
  sprintf (buffer, "%s%s%d", "/tmp/", scratch_filename, pid);
  if (options.verbosity >= 1)
    fprintf (stderr, "PicoSAT-stats scratch file: %s\n", buffer);

  FILE *scratch = fopen (buffer, "w+");
  if (!scratch)
    return decisions;

  picosat_set_output (scratch);
  picosat_stats ();
  fflush (scratch);
  rewind (scratch);

  char suffix[1024] = { 0 };
  int i = 0;
  /* Read at most 20 lines. */
  while (i < 20)
    {
      char line[1024] = { 0 };
      int tmp;
      fgets (line, 1023, scratch);
      sscanf (line, "c %d %s\n", &tmp, suffix);
      if (!strcmp (suffix, "decisions"))
        {
          decisions = tmp;
          break;
        }
      memset (suffix, 0, 1024);
      i++;
    }

  picosat_set_output (stderr);
  fclose (scratch);
  remove (buffer);
  return decisions;
}


/* Call SAT solver once to find a model which allows to rule out
   upcoming tests. */
static QxBFState
initial_sat_sweep_calls ()
{
  QxBFState state = STATE_UNDEF;
  if (options.sat_sweep_calls == 0)
    fprintf (stderr,
             "\nWARNING: skipping intial sat sweep calls! Using default bounds!\n\n");

  assert (options.sat_sweep_calls == 1);

  /* Initial unbounded sat calls to determine ignore-literals. */
  unsigned int cnt;
  for (cnt = 1; cnt <= options.sat_sweep_calls && state == STATE_UNDEF &&
       !time_limit_reached (&options.sat_time_limit, &total_sat_time); cnt++)
    {
      total_sat_tests++;
      double sat_start_time = time_stamp ();
      picosat_reset_phases ();

      if (options.sat_sweep_dec_limit)
        picosat_sat (options.sat_sweep_dec_limit);
      else
        picosat_sat (-1);

      sum_up_cur_time (&cur_sat_time_stamp, &total_sat_time);

      if (options.verbosity >= 1)
        {
          fprintf (stderr, "Initial PicoSAT Stats:\n");
          picosat_stats ();
        }

      double sat_time = time_stamp () - sat_start_time;
      if (options.verbosity >= 1)
        fprintf (stderr, "Initial sat-time: %f\n", sat_time);

      if (picosat_res () == PICOSAT_UNSATISFIABLE)
        state = STATE_UNSAT;
      else if (picosat_res () == PICOSAT_SATISFIABLE)
        {
          mark_ignore_vars_by_model (1);

          if (cnt == 1 && computed_dec_limit == 0)
            {
              computed_dec_limit = get_picosat_decisions ();
              if (computed_dec_limit == -1)
                {
                  QXBF_ABORT (1, "Parsing PicoSAT stats failed!");
                }
              else if (options.sat_time_threshold
                       && sat_time <= options.sat_time_threshold)
                {
                  /* Dynamic bound. */
                  computed_dec_limit += options.sat_dec_limit_inc;
                }
              else
                {
                  /* SAT-call too costly: take user-specified values or default ones. */
                  computed_dec_limit = options.init_sat_dec_limit =
                    options.sat_dec_limit;
                  if (options.verbosity >= 1)
                    fprintf (stderr,
                             "SAT-call too costly or no time-bound given, using static bounds.\n");
                }
              if (options.verbosity >= 1)
                fprintf (stderr, "Computed SAT decision-limit: %d\n",
                         computed_dec_limit);
              options.init_sat_dec_limit = options.sat_dec_limit =
                computed_dec_limit;
            }
        }
      else
        {
          assert (picosat_res () == PICOSAT_UNKNOWN);
          /* Was not able to solve SAT instance within decision
             limit. Treat as if time limit was exceeded. */
          computed_dec_limit = options.init_sat_dec_limit =
            options.sat_dec_limit;
          if (options.verbosity >= 1)
            fprintf (stderr,
                     "SAT-call aborted by decision limit, using static bounds.\n");
        }
    }

  sum_up_cur_time (&cur_sat_time_stamp, &total_sat_time);

  return state;
}


static QxBFState
sat_based_probing ()
{
  QxBFState state = STATE_UNDEF;
  cur_sat_time_stamp = time_stamp ();

  if (time_limit_reached (&options.sat_time_limit, &total_sat_time))
    return state;

  assert (next_to_propagate == assigned_vars_l1);

  picosat_init ();
  picosat_set_seed (options.seed);
  picosat_set_output (stderr);
  sat_based_probing_setup ();

  /* Initial-calls only before very first round of SAT-tests. */
  if (sat_rounds == 0 && (state = initial_sat_sweep_calls ()) == STATE_UNSAT)
    {
      /* Catch early UNSAT result. */
      cnf_unsat = 1;
      picosat_reset ();
      return state;
    }

  unsigned int top_level_size;
  int changed = 0;

  /* 'max_prop_bound' default 0, which means NO bound. */
  max_sat_prop_bound = options.sat_dec_limit;

  /* 'cur_prop_bound' in range [1,max_prop_bound], default 'max_prop_bound'. */
  cur_sat_prop_bound = options.init_sat_dec_limit > 0 ?
    (options.init_sat_dec_limit <
     max_sat_prop_bound ? options.
     init_sat_dec_limit : max_sat_prop_bound) : max_sat_prop_bound;

  do
    {
      assert (cur_sat_prop_bound <= max_sat_prop_bound);
      assert (cur_sat_prop_bound != 0 || max_sat_prop_bound == 0);
      assert (cur_sat_prop_bound == 0 || max_sat_prop_bound != 0);
      do
        {
          assert (next_to_propagate == assigned_vars_l1);
          /* Perform round with current propagation limit. */
          sat_rounds++;

          if (options.verbosity >= 1)
            fprintf (stderr, "SAT-Round %d, cur-bound %d, max-bound %d\n",
                     sat_rounds, cur_sat_prop_bound, max_sat_prop_bound);

          top_level_size = STACK_CNT (assigned_vars);
          state = sat_based_probing_aux ();
          assert (state != STATE_UNDEF
                  || next_to_propagate == assigned_vars_l1);
          changed =
            top_level_size != ((unsigned int) STACK_CNT (assigned_vars));

          if (options.verbosity >= 1)
            fprintf (stderr, "SAT-Round %d end, top-level change %d\n",
                     sat_rounds,
                     (int) (STACK_CNT (assigned_vars) - top_level_size));

          /* Changes to top-level assignment: perform another round
             with same propagation limit. */
          if (changed)
            if (opt_cur_sat_prop_bound < cur_sat_prop_bound)
              opt_cur_sat_prop_bound = cur_sat_prop_bound;
        }
      while (state == STATE_UNDEF && changed);

      if (cur_sat_prop_bound == max_sat_prop_bound)
        break;

      /* Increase current propagation limit, but bound to 'used_vars'. */
      cur_sat_prop_bound = cur_sat_prop_bound * 2;
      if (cur_sat_prop_bound > max_sat_prop_bound)
        cur_sat_prop_bound = max_sat_prop_bound;
    }
  while (state == STATE_UNDEF && cur_sat_prop_bound <= max_sat_prop_bound);

  picosat_reset ();

  sum_up_cur_time (&cur_sat_time_stamp, &total_sat_time);
  assert (state != STATE_UNDEF || next_to_propagate == assigned_vars_l1);

  return state;
}


static QxBFState
preprocess ()
{
  QxBFState state = STATE_UNDEF;
  /* Initialize time stamp for BCP start. */
  cur_bcp_time_stamp = time_stamp ();

  if (time_limit_reached (&options.bcp_time_limit, &total_bcp_time))
    return state;

  unsigned int top_level_size;
  int changed = 0;
  unsigned int used_vars = num_vars;

  /* 'max_prop_bound' in range [1,used_vars], default 'used_vars'. */
  max_bcp_prop_bound = options.bcp_prop_limit > 0 ?
    (options.bcp_prop_limit <
     used_vars ? options.bcp_prop_limit : used_vars) : used_vars;

  /* 'cur_prop_bound' in range [1,max_prop_bound], default 'max_prop_bound'. */
  cur_bcp_prop_bound = options.init_bcp_prop_limit > 0 ?
    (options.init_bcp_prop_limit <
     max_bcp_prop_bound ? options.
     init_bcp_prop_limit : max_bcp_prop_bound) : max_bcp_prop_bound;

  do
    {
      assert (0 < cur_bcp_prop_bound
              && cur_bcp_prop_bound <= max_bcp_prop_bound);
      do
        {
          /* Perform round with current propagation limit. */
          bcp_rounds++;
          if (options.verbosity >= 1)
            fprintf (stderr, "BCP-Round %d, cur-bound %d, max-bound %d\n",
                     bcp_rounds, cur_bcp_prop_bound, max_bcp_prop_bound);
          top_level_size = STACK_CNT (assigned_vars);
          state = bcp ();
          changed =
            top_level_size != ((unsigned int) STACK_CNT (assigned_vars));
          /* Changes to top-level assignment: perform another round
             with same propagation limit. */
          if (changed)
            if (opt_cur_bcp_prop_bound < cur_bcp_prop_bound)
              opt_cur_bcp_prop_bound = cur_bcp_prop_bound;
        }
      while (state == STATE_UNDEF && changed);

      if (cur_bcp_prop_bound == max_bcp_prop_bound)
        break;

      /* Increase current propagation limit, but bound to 'used_vars'. */
      cur_bcp_prop_bound = cur_bcp_prop_bound * 2;
      if (cur_bcp_prop_bound > max_bcp_prop_bound)
        cur_bcp_prop_bound = max_bcp_prop_bound;
    }
  while (state == STATE_UNDEF && cur_bcp_prop_bound <= max_bcp_prop_bound);

  sum_up_cur_time (&cur_bcp_time_stamp, &total_bcp_time);

  return state;
}


static void
print_config (int argc, char **argv)
{
  fprintf (stderr, "\n---------- CONFIG ---------- \n\n");
  int i;
  for (i = 1; i < argc; i++)
    fprintf (stderr, "argv[%d]: %s\n", i, argv[i]);
  fprintf (stderr, "\n");

  fprintf (stderr, "-s=%d\n", options.seed);
  fprintf (stderr, "--no-pures=%d\n", options.no_pures);
  fprintf (stderr, "--no-abs=%d\n", options.no_abs);
  fprintf (stderr, "--no-bcp=%d\n", options.no_bcp);
  fprintf (stderr, "--no-sat=%d\n", options.no_sat);
  fprintf (stderr, "--bcp-prop-limit=%d\n", options.bcp_prop_limit);
  fprintf (stderr, "--init-bcp-prop-limit=%d\n", options.init_bcp_prop_limit);
  fprintf (stderr, "--sat-dec-limit=%d\n", options.sat_dec_limit);
  fprintf (stderr, "--init-sat-dec-limit=%d\n", options.init_sat_dec_limit);
  fprintf (stderr, "--sat-ignore-dist=%d\n", options.sat_ignore_dist);
  fprintf (stderr, "--sat-sweep-calls=%d\n", options.sat_sweep_calls);
  fprintf (stderr, "--sat-sweep-dec-limit=%d\n", options.sat_sweep_dec_limit);
  fprintf (stderr, "--sat-dec-limit-inc=%d\n", options.sat_dec_limit_inc);
  fprintf (stderr, "--sat-time-threshold=%f\n", options.sat_time_threshold);
  fprintf (stderr, "--sat-time-limit=%d\n", options.sat_time_limit);
  fprintf (stderr, "--bcp-time-limit=%d\n", options.bcp_time_limit);

  fprintf (stderr, "\n------------------------------ \n\n");
}


static void
print_stats ()
{
  fprintf (stderr, "\n---------- STATS ---------- \n\n");
  fprintf (stderr, "QxBF time: %f\n", end_time - start_time);
  fprintf (stderr, "BCP-Rounds: %d\n", bcp_rounds);
  fprintf (stderr, "BCP-Time: %f\n", total_bcp_time);
  fprintf (stderr, "BCP-Units from failed lits: %d\n",
           learnt_from_failed_lits);
  fprintf (stderr, "Total BCP-decisions: %llu\n", total_bcp_decisions);
  fprintf (stderr, "Total BCP-propagations: %llu\n", total_bcp_propagations);
  fprintf (stderr, "Max. cur. BCP-propagations: %d\n",
           max_cur_bcp_propagations);
  fprintf (stderr, "Opt. cur. BCP-prop. bound: %d\n", opt_cur_bcp_prop_bound);
  double avg_bcp_prop =
    total_bcp_decisions ? total_bcp_propagations /
    (double) total_bcp_decisions : 0;
  fprintf (stderr, "Avg. BCP-propagations: %f (%f%% of max. %d)\n",
           avg_bcp_prop, (avg_bcp_prop / num_vars * 100), num_vars);
  fprintf (stderr, "\n");
  fprintf (stderr, "SAT-Rounds: %d\n", sat_rounds);
  fprintf (stderr, "SAT-Tests: %llu\n", total_sat_tests);
  fprintf (stderr, "SAT-Solver models: %u\n", total_sat_solver_models);
  fprintf (stderr, "Unsatisfiable CNF: %s\n", cnf_unsat ? "yes" : "no");
  fprintf (stderr, "SAT-Time: %f (avg. %f)\n", total_sat_time,
           total_sat_tests ? (total_sat_time / total_sat_tests) : 0);
  fprintf (stderr, "SAT-Units from UNSAT tests: %d\n",
           learnt_from_unsat_tests);
  fprintf (stderr, "SAT-Units from solver: %d\n", learnt_from_sat_solver);
  fprintf (stderr, "\n");
  fprintf (stderr, "Top-level size: %d\n",
           (int) (assigned_vars_l1 - assigned_vars_l0));
  fprintf (stderr, "\n------------------------------ \n\n");
}


static QxBFState
run_preprocessing ()
{
  int changed = 0;
  QxBFState state = STATE_UNDEF;

  do
    {
      assert (next_to_propagate == assigned_vars_l1);
      /* Do BCP-based preprocessing until saturation or limit is reached. */
      if (!options.no_bcp)
        state = preprocess ();
      assert (state != STATE_UNDEF || next_to_propagate == assigned_vars_l1);
      unsigned int top_level_size = STACK_CNT (assigned_vars);
      if (state == STATE_UNDEF)
        if (!options.no_sat)
          state = sat_based_probing ();
      changed = top_level_size != ((unsigned int) STACK_CNT (assigned_vars));
      assert (state != STATE_UNDEF || next_to_propagate == assigned_vars_l1);
    }
  while (state == STATE_UNDEF && changed);
  return state;
}


static int
isnumstr (char *str)
{
  /* Empty string is not considered as number-string. */
  if (!*str)
    return 0;
  char *p;
  for (p = str; *p; p++)
    {
      char c = *p;
      if (!isdigit (c))
        return 0;
    }
  return 1;
}


int
main (int argc, char **argv)
{
  QxBFState state = STATE_UNDEF;
  int i, j, res = 0;
  DIR *dir = NULL;
  char *str;
  start_time = time_stamp ();

  options.no_pures = 0;
  options.seed = 0;
  options.verbosity = 0;
  options.no_abs = 0;
  options.no_bcp = 0;
  options.no_sat = 0;
  options.bcp_prop_limit = 4096;
  options.init_bcp_prop_limit = 0;
  options.sat_dec_limit = 1000;
  options.init_sat_dec_limit = 1000;
  options.sat_ignore_dist = 1;
  options.sat_sweep_calls = 1;
  options.sat_sweep_dec_limit = 500000;
  options.sat_time_threshold = 1000 / 1000;
  options.sat_time_limit = 40;
  options.bcp_time_limit = 40;
  options.sat_dec_limit_inc = 8000;
  options.simple_output = 0;
  options.full_clean_output = 0;
  options.print_toplevel = 0;

  STACK_INIT (qres_wreason, 1);
  STACK_INIT (qres_antecedent, 1);
  STACK_INIT (qres_tmp, 1);

  /* Default: read from stdin. */
  in = stdin;

  for (i = 1; i < argc; i++)
    {
      str = argv[i];
      if (strcmp (str, "-h") == 0 || strcmp (str, "--help") == 0)
        {
          print_usage ();
          return (EXIT_SUCCESS);
        }
      else if (strcmp (str, "--version") == 0)
        {
          print_version ();
          return (EXIT_SUCCESS);
        }
      else if (strcmp (str, "-v") == 0)
        {
          options.verbosity++;
        }
      else if (strcmp (str, "-s") == 0)
        {
          i++;
          QXBF_ABORT (i >= argc, "missing value for option -s");
          str = argv[i];
          j = 0;
          while (argv[i][j] != '\0')
            QXBF_ABORT (!isdigit (argv[i][j++]),
                        "invalid value for option -s");
          options.seed = (unsigned int) atoi (argv[i]);
        }
      else if (strcmp (str, "--no-pures") == 0)
        {
          options.no_pures = 1;
        }
      else if (strcmp (str, "--no-abs") == 0)
        {
          options.no_abs = 1;
        }
      else if (strcmp (str, "--simple-output") == 0)
        {
          options.simple_output = 1;
        }
      else if (strcmp (str, "--full-clean-output") == 0)
        {
          options.full_clean_output = 1;
        }
      else if (strcmp (str, "--print-toplevel") == 0)
        {
          options.print_toplevel = 1;
        }
      else if (strcmp (str, "--no-bcp") == 0)
        {
          options.no_bcp = 1;
        }
      else if (strcmp (str, "--no-sat") == 0)
        {
          options.no_sat = 1;
        }
      else
        if (!strncmp (str, "--bcp-prop-limit=", strlen ("--bcp-prop-limit=")))
        {
          str += strlen ("--bcp-prop-limit=");
          QXBF_ABORT ((!isnumstr (str)),
                      "Expecting non-negative integer after '--bcp-prop-limit='");
          options.bcp_prop_limit = atoi (str);
        }
      else
        if (!strncmp
            (str, "--init-bcp-prop-limit=",
             strlen ("--init-bcp-prop-limit=")))
        {
          str += strlen ("--init-bcp-prop-limit=");
          QXBF_ABORT ((!isnumstr (str)),
                      "Expecting non-negative integer after '--init-bcp-prop-limit='");
          options.init_bcp_prop_limit = atoi (str);
        }
      else
        if (!strncmp (str, "--sat-dec-limit=", strlen ("--sat-dec-limit=")))
        {
          str += strlen ("--sat-dec-limit=");
          QXBF_ABORT ((!isnumstr (str)),
                      "Expecting non-negative integer after '--sat-dec-limit='");
          options.sat_dec_limit = atoi (str);
        }
      else
        if (!strncmp
            (str, "--init-sat-dec-limit=", strlen ("--init-sat-dec-limit=")))
        {
          str += strlen ("--init-sat-dec-limit=");
          QXBF_ABORT ((!isnumstr (str)),
                      "Expecting non-negative integer after '--init-sat-dec-limit='");
          options.init_sat_dec_limit = atoi (str);
        }
      else
        if (!strncmp
            (str, "--sat-ignore-dist=", strlen ("--sat-ignore-dist=")))
        {
          str += strlen ("--sat-ignore-dist=");
          QXBF_ABORT ((!isnumstr (str)),
                      "Expecting non-negative integer after '--sat-ignore-dist='");
          options.sat_ignore_dist = atoi (str);
          QXBF_ABORT ((options.sat_ignore_dist >= 32),
                      "Expecting integer in [0,31] for '--sat-ignore-dist='");
        }
      else
        if (!strncmp
            (str, "--sat-sweep-calls=", strlen ("--sat-sweep-calls=")))
        {
          str += strlen ("--sat-sweep-calls=");
          QXBF_ABORT ((!isnumstr (str)),
                      "Expecting non-negative integer after '--sat-sweep-calls='");
          options.sat_sweep_calls = atoi (str);
        }
      else
        if (!strncmp
            (str, "--sat-sweep-dec-limit=",
             strlen ("--sat-sweep-dec-limit=")))
        {
          str += strlen ("--sat-sweep-dec-limit=");
          QXBF_ABORT ((!isnumstr (str)),
                      "Expecting non-negative integer after '--sat-sweep-dec-limit='");
          options.sat_sweep_dec_limit = atoi (str);
        }
      else
        if (!strncmp
            (str, "--sat-time-threshold=", strlen ("--sat-time-threshold=")))
        {
          str += strlen ("--sat-time-threshold=");
          QXBF_ABORT ((!isnumstr (str)),
                      "Expecting non-negative integer after '--sat-time-threshold='");
          options.sat_time_threshold = ((double) atoi (str)) / 1000;
        }
      else
        if (!strncmp (str, "--sat-time-limit=", strlen ("--sat-time-limit=")))
        {
          str += strlen ("--sat-time-limit=");
          QXBF_ABORT ((!isnumstr (str)),
                      "Expecting non-negative integer after '--sat-time-limit='");
          options.sat_time_limit = atoi (str);
        }
      else
        if (!strncmp (str, "--bcp-time-limit=", strlen ("--bcp-time-limit=")))
        {
          str += strlen ("--bcp-time-limit=");
          QXBF_ABORT ((!isnumstr (str)),
                      "Expecting non-negative integer after '--bcp-time-limit='");
          options.bcp_time_limit = atoi (str);
        }
      else
        if (!strncmp
            (str, "--sat-dec-limit-inc=", strlen ("--sat-dec-limit-inc=")))
        {
          str += strlen ("--sat-dec-limit-inc=");
          QXBF_ABORT ((!isnumstr (str)),
                      "Expecting non-negative integer after '--sat-dec-limit-inc='");
          options.sat_dec_limit_inc = atoi (str);
        }
      else if (str[0] == '-')
        QXBF_ABORT (1, "unknown option");
      else
        {
          in_filename = str;
          if ((dir = opendir (str)) != NULL)
            {
              closedir (dir);
              QXBF_ABORT (1, "given file is a directory");
            }

          in = fopen (str, "r");
          QXBF_ABORT (in == NULL, "failed to open input file");
        }
    }

  QXBF_ABORT ((options.sat_dec_limit != options.init_sat_dec_limit),
              "Expecting equal values for --sat-dec-limit and --init-sat-dec-limit!\n");
  QXBF_ABORT ((options.sat_sweep_calls != 1),
              "Expecting exactly one sweep call!\n");
  QXBF_ABORT ((options.simple_output
               && options.full_clean_output),
              "Contradicting output options!\n");

  if (options.seed)
    srand (options.seed);
  else
    srand (time (NULL) * getpid ());

  if (options.verbosity >= 1)
    print_config (argc, argv);

  state = parse ();

#ifndef NDEBUG
  assert_check_occurences ();
  assert_check_var_scope_levels ();
#endif

  if (state == STATE_UNDEF)
    {
      next_to_propagate = assigned_vars.bp;
      state = level0_bcp ();

      if (state == STATE_UNDEF)
        state = run_preprocessing ();
    }

#ifndef NDEBUG
  assert_check_pure_literals (__LINE__);
#endif

  if (state == STATE_UNSAT)
    {
      res = 20;
      fprintf (stderr, "QxBF result: UNSAT\n");
    }
  else if (state == STATE_SAT)
    {
      res = 10;
      fprintf (stderr, "QxBF result: SAT\n");
    }
  else if (state == STATE_UNDEF)
    {
      res = 0;
      fprintf (stderr, "QxBF result: UNDEFINED\n");
    }

  if (options.print_toplevel)
    print_sorted_toplevel ();

  if (options.simple_output)
    print_result_append (stdout, state);
  else if (options.full_clean_output)
    print_result_full_clean (stdout, state);
  else
    print_result_clean (stdout, state);

  cleanup ();

  fclose (in);
  end_time = time_stamp ();

  if (options.verbosity >= 1)
    print_stats ();

  exit (res);
}


#ifndef NDEBUG

static void
dump (void)
{
  int *tmp;

  for (tmp = assigned_vars_l0; tmp != assigned_vars_l0 + assigned_vars.size;
       tmp++)
    {
      fprintf (stderr, "level: %d ", vars[*tmp].d_level);
      fprintf (stderr, "var: %8d (%d) mode: %d, occs: +(%d/%d) -(%d/%d)\n",
               *tmp, vars[*tmp].assignment, vars[*tmp].mode,
               vars[*tmp].num_pos_occs, vars[*tmp].num_pos_occs_cur,
               vars[*tmp].num_neg_occs, vars[*tmp].num_neg_occs_cur);
    }
  fprintf (stderr, "\n");
}

static void
assert_check_pure_literals (int line)
{
  int i;

  for (i = 1; i < num_vars; i++)
    {
      if (vars[i].mode == VMODE_PURE)
        {
          if (!((vars[i].num_pos_occs_cur == 0 &&
                 vars[i].num_neg_occs_cur >= 0) ||
                (vars[i].num_neg_occs_cur == 0 &&
                 vars[i].num_pos_occs_cur >= 0)))
            {
              dump ();
              fprintf (stderr, "line: %d\n", line);
            }

          assert ((vars[i].num_pos_occs_cur == 0 &&
                   vars[i].num_neg_occs_cur >= 0) ||
                  (vars[i].num_neg_occs_cur == 0 &&
                   vars[i].num_pos_occs_cur >= 0));

          assert (vars[i].type == QTYPE_EXISTS
                  || vars[i].type == QTYPE_FORALL);

          if (vars[i].type == QTYPE_EXISTS)
            {
              if (vars[i].num_pos_occs_cur == 0
                  && vars[i].num_neg_occs_cur > 0)
                assert (vars[i].assignment == BTYPE_FALSE);
              else if (vars[i].num_neg_occs_cur == 0
                       && vars[i].num_pos_occs_cur > 0)
                assert (vars[i].assignment == BTYPE_TRUE);
            }
        }
    }
}


static void
assert_check_occurences (void)
{
  int j, k, num_occs_tmp;
  ClauseId i, *occs_tmp;
  Lit lit;

  for (i = 0; i < (ClauseId) num_clauses; i++)
    {
      for (j = 0; j < clauses[i].num_lits; j++)
        {
          lit = clauses[i].lits[j];
          if (lit < 0)
            {
              num_occs_tmp = vars[-lit].num_neg_occs;
              occs_tmp = vars[-lit].neg_occs;
            }
          else
            {
              num_occs_tmp = vars[lit].num_pos_occs;
              occs_tmp = vars[lit].pos_occs;
            }

          for (k = 0; k < num_occs_tmp - 1; k++)
            if (occs_tmp[k] == i)
              break;

          assert (occs_tmp[k] == i);
        }
    }
}


static void
assert_check_var_scope_levels (void)
{
  int i, scope;

  for (scope = outermost_scope_index; scope < num_scopes; scope++)
    for (i = 0; i < scopes[scope].num_vars; i++)
      assert (vars[scopes[scope].vars[i]].s_level == scope);
}


static void
assert_check_is_propagated (void)
{
  int *tmp;
  assert (assigned_vars_l0 != NULL);
  assert (assigned_vars_l1 != NULL);
  for (tmp = assigned_vars_l0; tmp < assigned_vars_l1; tmp++)
    assert (vars[*tmp].is_propagated);
}
#endif
